import math
import pickle
import sys
from decimal import Decimal
from fractions import Fraction

import pytest
from hypothesis import example, given, settings
from hypothesis.strategies import floats
from supportclasses import a, b, c, d, q, r, z

import gmpy2
from gmpy2 import (cmp, cmp_abs, from_binary, gamma_inc, get_context, inf,
                   is_nan, mpc, mpfr, mpfr_grandom, mpfr_nrandom, mpq, mpz,
                   nan, random_state, to_binary, xmpz, zero)


def test_mpfr_gamma_inc():
    assert gamma_inc(1, 1) == mpfr('0.36787944117144233')
    assert gamma_inc(1, 0) == mpfr('1.0')
    assert gamma_inc(0, 1) == mpfr('0.21938393439552029')


def test_mpfr_cmp():
    assert cmp(mpfr(0), mpfr(0)) == 0
    assert cmp(mpfr(0), mpz(0)) == 0
    assert cmp(mpfr(0), mpq(0,1)) == 0
    assert cmp(zero(-1), zero(-1)) == 0
    assert cmp(zero(1), zero(-1)) == 0
    assert cmp(zero(-1), zero(1)) == 0
    assert cmp(mpfr(1), mpfr(0)) == 1
    assert cmp(mpfr(1), mpz(0)) == 1
    assert cmp(mpfr(1), mpq(0,1)) == 1
    assert cmp(mpfr(-1), mpfr(0)) == -1
    assert cmp(mpfr(-1), mpz(0)) == -1
    assert cmp(mpfr(-1), mpq(0,1)) == -1
    assert cmp(nan(), mpfr(0)) == 0
    assert cmp(nan(), mpz(0)) == 0
    assert cmp(nan(), mpq(0,1)) == 0

    gmpy2.get_context().clear_flags()

    assert cmp(nan(), 1) == 0
    assert gmpy2.get_context().erange is True

    assert cmp_abs(mpfr(-1), mpfr(0)) == 1
    assert cmp_abs(mpfr(-1), mpz(0)) == 1
    assert cmp_abs(mpfr(-1), mpq(0,1)) == 1
    assert cmp_abs(mpfr(0), mpfr(-1)) == -1
    assert cmp_abs(mpz(0), mpfr(-1)) == -1
    assert cmp_abs(mpq(0,1), mpfr(-1)) == -1

    assert cmp(mpfr(1.5), q) == 0
    assert cmp(r, mpfr(1.5)) == 0


def test_mpfr_comparisons():
    from supportclasses import a, q, r

    assert mpfr(1.5) == q
    assert r == mpfr(1.5)
    assert (r == a) is False

    a = mpz(123)
    r = mpfr('inf')
    q = mpq('45/7')
    f = float(0.7)
    r2 = mpfr(454.6)

    assert (r == a, r != a, r > a, r >= a, r < a, r <= a) == (False, True, True, True, False, False)

    r = mpfr('-inf')

    assert (r == a, r != a, r > a, r >= a, r < a, r <= a) == (False, True, False, False, True, True)

    r = mpfr('nan')

    assert (r == a, r != a, r > a, r >= a, r < a, r <= a) == (False, True, False, False, False, False)
    assert (r == q, r != q, r > q, r >= q, r < q, r <= q) == (False, True, False, False, False, False)
    assert (r == f, r != f, r > f, r >= f, r < f, r <= f) == (False, True, False, False, False, False)
    assert (r == r2, r != r2, r > r2, r >= r2, r < r2, r <= r2) == (False, True, False, False, False, False)

    r = mpfr(126.5)

    assert (r == a, r != a, r > a, r >= a, r < a, r <= a) == (False, True, True, True, False, False)
    assert (r == q, r != q, r > q, r >= q, r < q, r <= q) == (False, True, True, True, False, False)

    f = float(126.5)

    assert (r == f, r != f, r > f, r >= f, r < f, r <= f) == (True, False, False, True, False, True)
    assert (r == r2, r != r2, r > r2, r >= r2, r < r2, r <= r2) == (False, True, False, False, True, True)


def test_mpfr_conversion():
    x = mpfr(a)
    assert isinstance(x, mpfr)
    assert x == 1.5
    pytest.raises(TypeError, lambda: mpfr(b))
    pytest.raises(TypeError, lambda: mpfr(c))
    pytest.raises(TypeError, lambda: mpfr(d))

    pytest.raises(OverflowError, lambda: mpz(mpfr('inf')))

    assert mpz(mpfr(5.51)) == mpz(6)

    pytest.raises(OverflowError, lambda: xmpz(mpfr('inf')))

    assert xmpz(mpfr(5.51)) == xmpz(6)

    pytest.raises(OverflowError, lambda: mpq(mpfr('inf')))

    assert mpq(mpfr(4.5)) == mpq(9,2)
    assert mpq(mpfr(0)) == mpq(0,1)
    assert int(mpfr(5.3)) == 5
    assert float(mpfr(5.3)) == 5.3
    assert str(float('5.656')) == '5.656'


def test_mpfr_create():
    assert mpfr() == mpfr('0.0')
    assert mpfr(0) == mpfr('0.0')
    assert mpfr(1) == mpfr('1.0')
    assert mpfr(-1) == mpfr('-1.0')
    assert mpfr("123e17") == mpfr('1.23e+19')
    assert mpfr('1._3_5e4_5') == mpfr('1.3499999999999999e+45')

    pytest.raises(ValueError, lambda: mpfr("foo"))

    assert mpfr(1,100) == mpfr('1.0',100)
    assert mpfr("1",100) == mpfr('1.0',100)

    pytest.raises(TypeError, lambda: mpfr("1","hi"))

    assert mpfr("-inf") == mpfr('-inf')
    assert mpfr("inf") == mpfr('inf')
    assert is_nan(mpfr("nan"))
    assert mpfr(float("-inf")) == mpfr('-inf')
    assert mpfr(float("inf")) == mpfr('inf')
    assert is_nan(mpfr(float("nan")))
    assert mpfr(float("0")) == mpfr('0.0')
    assert mpfr(float("-0")) == mpfr('-0.0')
    assert mpfr("-0") == mpfr('-0.0')

    a = float("1.2345678901234567890")
    b = mpfr("1.2345678901234567890")
    assert a == 1.2345678901234567
    assert b == mpfr('1.2345678901234567')
    assert a == b

    c = mpfr(b)

    assert b is c
    assert mpfr(Fraction(0,1)) == mpfr('0.0')
    assert mpfr(Fraction(0,-1)) == mpfr('0.0')
    assert mpfr(Fraction(1,2)) == mpfr('0.5')
    assert mpfr(-1) == mpfr('-1.0')
    assert mpfr(12345678901234567890) == mpfr('1.2345678901234567e+19')
    assert mpfr(mpz(12345678901234567890)) == mpfr('1.2345678901234567e+19')
    assert mpfr(mpz(-1)) == mpfr('-1.0')
    assert mpfr(2**15 - 1) == mpfr('32767.0')
    assert mpfr(2**15) == mpfr('32768.0')
    assert mpfr(2**15 + 1) == mpfr('32769.0')
    assert mpfr(2**30 - 1) == mpfr('1073741823.0')
    assert mpfr(2**30 ) == mpfr('1073741824.0')
    assert mpfr(2**30 + 1) == mpfr('1073741825.0')

    ctx = gmpy2.get_context()
    ctx.clear_flags()
    a = mpfr("1.2")

    assert a.rc == -1

    ctx.clear_flags()
    a = mpfr("1.25")

    assert a.rc == 0

    ctx.clear_flags()
    a = mpfr('nan')

    assert is_nan(a)
    assert ctx.invalid

    ctx.clear_flags()

    assert is_nan(mpfr(a))
    assert not ctx.invalid

    ctx.clear_flags()

    assert is_nan(mpfr(float('nan')))
    assert ctx.invalid

    gmpy2.set_context(gmpy2.ieee(128))

    assert mpfr(1)/7 == mpfr('0.14285714285714285714285714285714285',113)
    assert mpfr(1.0/7) == mpfr('0.142857142857142849212692681248881854',113)
    assert mpfr(1.0/7).digits(2) == ('10010010010010010010010010010010010010010010010010010000000000000000000000000000000000000000000000000000000000000', -2, 113)

    gmpy2.set_context(gmpy2.ieee(32))

    assert mpfr(1)/7 == mpfr('0.142857149',24)
    assert (mpfr(1)/7).digits(2) == ('100100100100100100100101', -2, 24)
    assert mpfr(1.0/7) == mpfr('0.142857149',24)
    assert mpfr(1.0/7).digits(2) == ('100100100100100100100101', -2, 24)
    assert mpfr(1.0/7, precision=0) == mpfr('0.142857149',24)
    assert repr(mpfr(1.0/7, precision=1)) == "mpfr('0.14285714285714285')"
    assert repr(mpfr(1.0/7, precision=5)) == "mpfr('0.141',5)"


@settings(max_examples=1000)
@given(floats())
@example(0.0)
@example(1.0)
@example(2.0)
@example(-1.0)
@example(123.456)
@example(123.5)
@example(float('inf'))
@example(-float('inf'))
@example(float('nan'))
def test_mpfr_hash(x):
    if math.isnan(x):
        if sys.version_info < (3, 10):
            assert hash(mpfr(x)) == hash(x) == sys.hash_info.nan
        else:
            assert hash(mpfr(x)) != hash(x)
    else:
        assert hash(mpfr(x)) == hash(x)


@given(floats())
@example(0.0)
@example(1.0)
@example(-1.0)
@example(+float('inf'))
@example(-float('inf'))
@example(float('nan'))
@example(1.345)
def test_mpfr_to_from_binary_bulk(r):
    x = mpfr(r)
    y = from_binary(to_binary(x))
    assert x == y or (is_nan(x) and is_nan(y))


def test_mpfr_to_from_binary():
    x = mpfr("1.345e1000")
    assert x == from_binary(to_binary(x))
    assert -x == from_binary(to_binary(-x))
    x = mpfr("1e1234567890123456789")
    assert x == from_binary(to_binary(x))
    assert x.rc == 1
    x = mpfr("-1e1234567890123456789")
    assert x == from_binary(to_binary(x))
    assert x.rc == -1
    x = gmpy2.const_pi()
    assert x.rc == -1
    y = from_binary(to_binary(x))
    assert x == y and y.rc == -1
    -1
    x = gmpy2.const_pi(precision=104)
    assert x.rc == 1
    y = from_binary(to_binary(x))
    assert x == y and y.rc == 1
    with gmpy2.context() as ctx:
        ctx.precision = 20
        x = gmpy2.const_pi()
        assert x == from_binary(to_binary(x))
        ctx.precision = 100
        x = gmpy2.const_pi()
        assert x == from_binary(to_binary(x))
        ctx.precision = 200
        x = mpfr(gmpy2.const_pi())
        assert x == from_binary(to_binary(x))
        x = gmpy2.const_pi()
        ctx.precision = 300
        x = from_binary(to_binary(x))
        assert x.precision == 200


def test_mpfr_random():
    assert gmpy2.mpfr_random(random_state(42)) == mpfr('0.93002690534702315')


def test_mpfr_grandom():
    assert mpfr_grandom(random_state(42)) == (mpfr('-0.32898912492644183'),
                                              mpfr('0.03656576719642516'))


def test_mpfr_nrandom():
    assert mpfr_nrandom(random_state(42)) == mpfr('-0.32898912492644183')


def test_mpfr_mpmath():
    mpmath = pytest.importorskip("mpmath")
    a, b, c, d = '1.1', '-1.1', '-3.14', '0'
    assert mpfr(a)._mpf_ == (0, mpz(4953959590107546), -52, 53)
    assert mpmath.mpf(mpfr(a)) == mpmath.mpf(a)
    assert mpfr(b)._mpf_ == (1, mpz(4953959590107546), -52, 53)
    assert mpmath.mpf(mpfr(b)) == mpmath.mpf(b)
    assert mpfr(c, precision=10)._mpf_ == (1, mpz(804), -8, 10)
    assert mpmath.mpf(mpfr(c, precision=10), prec=10) == mpmath.mpf(c, prec=10)
    assert mpfr(d)._mpf_ == (0, mpz(0), 1, 1)
    pytest.raises(TypeError, lambda: gmpy2._mpmath_create(1))
    pytest.raises(ValueError, lambda: gmpy2._mpmath_create(1, 1, -1))
    pytest.raises(TypeError, lambda: mpmath.mpf(("!", 1,)))


def test_mpfr_format():
    r, r1, r2 = mpfr(5.6), mpfr(-3), mpfr(5)

    assert '{:<30}'.format(r1) == '-3.000000                     '
    assert '{:>+20}'.format(r2) == '           +5.000000'
    assert '{:>-15}'.format(r2) == '       5.000000'
    assert '{:>-15}'.format(r1) == '      -3.000000'
    assert '{:U}'.format(r) == '5.600000'

    pytest.raises(ValueError, lambda: '{:U-}'.format(r))

    assert '{:Z}'.format(r) == '5.599999'

    pytest.raises(ValueError, lambda: '{:Z+}'.format(r))

    assert '{:+Z}'.format(r) == '+5.599999'

    pytest.raises(ValueError, lambda: '{:Z }'.format(r))

    assert '{:.10f}'.format(r) == '5.6000000000'
    assert '{:.10f.}'.format(r) == '5.6000000000'

    assert "{:.0f}".format(mpfr('123')) == '123.0'

    pytest.raises(ValueError, lambda: '{:Z.}'.format(r))
    pytest.raises(ValueError, lambda: '{:->}'.format(r))
    pytest.raises(ValueError, lambda: '{:YZ}'.format(r))


def test_mpfr_digits():
    r, r2 = mpfr(5.6), mpfr(5)

    assert r.digits() == ('55999999999999996', 1, 53)
    assert r.digits(2) == ('10110011001100110011001100110011001100110011001100110', 3, 53)
    assert r.digits(2,54) == ('101100110011001100110011001100110011001100110011001100', 3, 53)
    assert r.digits(10,54) == ('559999999999999964472863211994990706443786621093750000', 1, 53)
    assert r2.digits(2) == ('10100000000000000000000000000000000000000000000000000', 3, 53)

    pytest.raises(TypeError, lambda: r2.digits(2, 5, 6))
    pytest.raises(ValueError, lambda: r.digits(0))


def test_mpfr_abs():
    a = mpfr(1.0)
    b = abs(a)

    assert a is not b
    assert abs(mpfr(1, precision=100)) == mpfr('1.0')

    ctx = gmpy2.get_context()
    ctx.clear_flags()

    assert is_nan(abs(mpfr('nan')))
    assert ctx.invalid

    ctx.clear_flags()

    assert abs(mpfr('inf')) == mpfr('inf')
    assert abs(mpfr('-inf')) == mpfr('inf')


def test_mpfr_not():
    assert not mpfr(0)
    assert not mpfr(1) is False


def test_mpfr_add():
    a = mpfr("12.34")
    b = mpfr("45.67")

    assert a+1 == mpfr('13.34')
    assert a+1.0 == mpfr('13.34')
    assert a+mpz(1) == mpfr('13.34')
    assert a+mpq(1,1) == mpfr('13.34')
    assert 1+a == mpfr('13.34')
    assert 1.0+a == mpfr('13.34')
    assert mpz(1)+a == mpfr('13.34')
    assert mpq(1,1)+a == mpfr('13.34')
    assert a+b == mpfr('58.010000000000005')
    assert a+0 == mpfr('12.34')
    assert 0+a == mpfr('12.34')
    assert b+a == mpfr('58.010000000000005')

    pytest.raises(TypeError, lambda: a+'b')
    pytest.raises(TypeError, lambda: 'b'+a)

    assert a == 12.34
    assert b == 45.67
    assert a+b == 12.34+45.67

    assert a + float('Inf') == mpfr('inf')
    assert float('Inf') + a == mpfr('inf')
    assert a + float('-Inf') == mpfr('-inf')
    assert float('-Inf') + a == mpfr('-inf')
    assert is_nan(a + float('nan'))
    assert is_nan(float('nan') + a)

    assert a + mpfr('Inf') == mpfr('inf')
    assert mpfr('Inf') + a == mpfr('inf')
    assert a + mpfr('-Inf') == mpfr('-inf')
    assert mpfr('-Inf') + a == mpfr('-inf')
    assert is_nan(a + mpfr('nan'))
    assert is_nan(mpfr('nan') + a)


def test_mpfr_sub():
    assert mpfr(10) - 1 == mpfr('9.0')
    assert 10 - mpfr(1) == mpfr('9.0')
    assert mpfr(10) - mpz(1) == mpfr('9.0')
    assert mpz(10) - mpfr(1) == mpfr('9.0')
    assert mpfr(10) - mpfr(1) == mpfr('9.0')
    assert mpfr(10) - mpq(1,1) == mpfr('9.0')
    assert mpq(10,1) - mpfr(1) == mpfr('9.0')
    assert mpfr(10) - Fraction(1,1) == mpfr('9.0')
    assert Fraction(10,1) - mpfr(1) == mpfr('9.0')
    assert mpfr(10) - 1.0 == mpfr('9.0')
    assert 10.0 - mpfr(1) == mpfr('9.0')
    assert mpfr(0) - (1 << 100) == mpfr('-1p100', base=2)
    assert (1 << 100) - mpfr(0) == mpfr('1p100', base=2)
    assert mpfr(10) - z == mpfr('8.0')
    assert mpfr(10) - q == mpfr('8.5')
    assert mpfr(10) - r == mpfr('8.5')

    a = mpfr("12.34")
    b = mpfr("45.67")

    assert a-1 == mpfr('11.34')
    assert a-1.0 == mpfr('11.34')
    assert a-(-1) == mpfr('13.34')
    assert a-(-1.0) == mpfr('13.34')
    assert a-b == mpfr('-33.329999999999998')
    assert b-a == mpfr('33.329999999999998')

    pytest.raises(TypeError, lambda: a-'b')
    pytest.raises(TypeError, lambda: 'b'-a)

    assert a-b == 12.34-45.67

    assert a - float('Inf') == mpfr('-inf')
    assert float('Inf') - a == mpfr('inf')
    assert a - float('-Inf') == mpfr('inf')
    assert float('-Inf') - a == mpfr('-inf')
    assert is_nan(a - float('nan'))
    assert is_nan(float('nan') - a)

    assert a - mpfr('Inf') == mpfr('-inf')
    assert mpfr('Inf') - a == mpfr('inf')
    assert a - mpfr('-Inf') == mpfr('inf')
    assert mpfr('-Inf') - a == mpfr('-inf')
    assert is_nan(a - mpfr('nan'))
    assert is_nan(mpfr('nan') - a)


def test_mpfr_mul():
    c = 12345678901234567890

    assert mpfr(10) * 1 == mpfr('10.0')
    assert 10 * mpfr(1) == mpfr('10.0')
    assert mpfr(10) * mpz(1) == mpfr('10.0')
    assert mpz(10) * mpfr(1) == mpfr('10.0')
    assert mpfr(10) * mpfr(1) == mpfr('10.0')
    assert mpfr(10) * mpq(1,1) == mpfr('10.0')
    assert mpq(10,1) * mpfr(1) == mpfr('10.0')
    assert mpfr(10) * Fraction(1,1) == mpfr('10.0')
    assert Fraction(10,1) * mpfr(1) == mpfr('10.0')
    assert mpfr(10) * 1.0 == mpfr('10.0')
    assert 10.0 * mpfr(1) == mpfr('10.0')
    assert mpfr(1) * c == mpfr(c)
    assert c * mpfr(1) == mpfr(c)
    assert mpfr(10) * z == mpfr('20.0')
    assert mpfr(10) * q == mpfr('15.0')
    assert mpfr(10) * r == mpfr('15.0')

    pytest.raises(TypeError, lambda: mpfr(10) * 'a')
    pytest.raises(TypeError, lambda: 'a' * mpfr(10))

    a = mpfr("12.34")
    b = mpfr("45.67")

    assert a*b == mpfr('563.56780000000003')
    assert a*0 == mpfr('0.0')
    assert a*0.0 == mpfr('0.0')
    assert a*1 == mpfr('12.34')
    assert a*1.0 == mpfr('12.34')
    assert a*(-1.0) == mpfr('-12.34')
    assert 0*a == mpfr('0.0')
    assert 0.0*a == mpfr('0.0')
    assert 1*a == mpfr('12.34')
    assert 1.0*a == mpfr('12.34')
    assert a*b == mpfr('563.56780000000003')
    assert a*b == 12.34*45.67

    assert a * float('Inf') == mpfr('inf')
    assert float('Inf') * a == mpfr('inf')
    assert a * float('-Inf') == mpfr('-inf')
    assert float('-Inf') * a == mpfr('-inf')
    assert -a * float('Inf') == mpfr('-inf')
    assert float('Inf') * -a == mpfr('-inf')
    assert -a * float('-Inf') == mpfr('inf')
    assert float('-Inf') * -a == mpfr('inf')
    assert is_nan(a * float('nan'))
    assert is_nan(float('nan') * a)
    assert is_nan(mpfr(0) * float('Inf'))
    assert is_nan(mpfr(0) * float('-Inf'))
    assert is_nan(float('Inf') * mpfr(0))
    assert is_nan(float('-Inf') * mpfr(0))

    assert a * mpfr('Inf') == mpfr('inf')
    assert mpfr('Inf') * a == mpfr('inf')
    assert a * mpfr('-Inf') == mpfr('-inf')
    assert mpfr('-Inf') * a == mpfr('-inf')
    assert -a * mpfr('Inf') == mpfr('-inf')
    assert mpfr('Inf') * -a == mpfr('-inf')
    assert -a * mpfr('-Inf') == mpfr('inf')
    assert mpfr('-Inf') * -a == mpfr('inf')
    assert is_nan(a * mpfr('nan'))
    assert is_nan(mpfr('nan') * a)
    assert is_nan(mpz(0) * mpfr('Inf'))
    assert is_nan(mpz(0) * mpfr('-Inf'))
    assert is_nan(mpfr('Inf') * mpfr(0))
    assert is_nan(mpfr('-Inf') * mpfr(0))


def test_mpfr_divmod():
    pytest.raises(TypeError, lambda: divmod(mpfr(1),'a'))

    ctx = gmpy2.context()

    assert ctx.divmod(mpfr(2),mpfr(1)) == (mpfr('2.0'), mpfr('0.0'))

    pytest.raises(TypeError, lambda: divmod(mpfr(1), mpc(1,2)))

    assert divmod(mpfr(1.2), mpfr(0.7)) == (mpfr('1.0'), mpfr('0.5'))

    ctx = gmpy2.ieee(64)
    gmpy2.set_context(ctx)

    assert divmod(mpfr(1.2), mpfr(0.7)) == (mpfr('1.0'), mpfr('0.5'))
    ctx.clear_flags()

    assert ctx.divzero is False
    assert all(map(ctx.is_nan, divmod(mpfr(1.2), mpfr(0))))

    with gmpy2.context(trap_divzero=True):
        pytest.raises(gmpy2.DivisionByZeroError, lambda: divmod(mpfr(1), mpfr(0)))
    with gmpy2.context(trap_invalid=True):
        pytest.raises(gmpy2.InvalidOperationError, lambda: divmod(gmpy2.nan(), mpfr(1)))
    with gmpy2.context(trap_invalid=True):
        pytest.raises(gmpy2.InvalidOperationError, lambda: divmod(mpfr(1), gmpy2.inf()))

    assert divmod(mpfr(111), mpfr(-222)) == (mpfr('-1.0'), mpfr('-111.0'))

    a = mpfr("12.34")
    b = mpfr("45.67")

    assert divmod(12.34, 45.67) == (0.0, 12.34)
    assert divmod(a,b) == (mpfr('0.0'), mpfr('12.34'))
    assert divmod(b,a) == (mpfr('3.0'), mpfr('8.6500000000000021'))
    assert divmod(45.67,12.34) == divmod(b,a)

    assert divmod(a, float('Inf')) == (mpfr('0.0'), mpfr('12.34'))
    assert divmod(a, float('-Inf')) == (mpfr('-1.0'), mpfr('-inf'))
    assert divmod(-a, float('Inf')) == (mpfr('-1.0'), mpfr('inf'))
    assert divmod(-a, float('-Inf')) == (mpfr('0.0'), mpfr('-12.34'))
    assert all(map(is_nan, divmod(a, float('nan'))))
    assert all(map(is_nan, divmod(-a, float('nan'))))
    assert divmod(mpfr(0), float('Inf')) == (mpfr('0.0'), mpfr('0.0'))
    assert divmod(mpfr(0), float('-Inf')) == (mpfr('-0.0'), mpfr('-0.0'))
    assert divmod(mpfr(0), float('nan'))
    assert all(map(is_nan, divmod(float('Inf'), a)))
    assert all(map(is_nan, divmod(float('-Inf'), a)))
    assert all(map(is_nan, divmod(float('Inf'), -a)))
    assert all(map(is_nan, divmod(float('-Inf'), -a)))
    assert all(map(is_nan, divmod(float('nan'), a)))
    assert all(map(is_nan, divmod(float('nan'), -a)))
    assert all(map(is_nan, divmod(float('Inf'), mpfr(0))))
    assert all(map(is_nan, divmod(float('-Inf'), mpfr(0))))
    assert all(map(is_nan, divmod(float('nan'), mpfr(0))))

    assert divmod(a, mpfr('Inf')) == (mpfr('0.0'), mpfr('12.34'))
    assert divmod(a, mpfr('-Inf')) == (mpfr('-1.0'), mpfr('-inf'))
    assert divmod(-a, mpfr('Inf')) == (mpfr('-1.0'), mpfr('inf'))
    assert divmod(-a, mpfr('-Inf')) == (mpfr('0.0'), mpfr('-12.34'))
    assert all(map(is_nan, divmod(a, mpfr('nan'))))
    assert all(map(is_nan, divmod(-a, mpfr('nan'))))
    assert divmod(mpfr(0), mpfr('Inf')) == (mpfr('0.0'), mpfr('0.0'))
    assert divmod(mpfr(0), mpfr('-Inf')) == (mpfr('-0.0'), mpfr('-0.0'))
    assert divmod(mpfr(0), mpfr('nan'))
    assert all(map(is_nan, divmod(mpfr('Inf'), a)))
    assert all(map(is_nan, divmod(mpfr('-Inf'), a)))
    assert all(map(is_nan, divmod(mpfr('Inf'), -a)))
    assert all(map(is_nan, divmod(mpfr('-Inf'), -a)))
    assert all(map(is_nan, divmod(mpfr('nan'), a)))
    assert all(map(is_nan, divmod(mpfr('nan'), -a)))
    assert all(map(is_nan, divmod(mpfr('Inf'), mpfr(0))))
    assert all(map(is_nan, divmod(mpfr('-Inf'), mpfr(0))))
    assert all(map(is_nan, divmod(mpfr('nan'), mpfr(0))))


def test_mpfr_floordiv():
    ctx = gmpy2.get_context()
    a, b = mpz(45), mpz(6)
    r, r2 = mpfr(45), mpfr(3.1)
    q, q2 = mpq(118,18), mpq(3,2)
    pyq, pyq2 = Fraction(118,18), Fraction(3,2)
    c, c2 = mpc(51, 65), mpc(4, 6)

    assert ctx.floor_div(r, r2) == mpfr('14.0')
    assert ctx.floor_div(r, r2) == mpfr('14.0')
    assert ctx.floor_div(r, 3.1) == mpfr('14.0')
    assert ctx.floor_div(r, 4) == mpfr('11.0')
    assert ctx.floor_div(r, b) == mpfr('7.0')
    assert ctx.floor_div(r, q2) == mpfr('30.0')
    assert ctx.floor_div(r, pyq2) == mpfr('30.0')
    assert ctx.floor_div(r, 0) == mpfr('inf')
    assert ctx.floor_div(r, mpz(0)) == mpfr('inf')
    assert ctx.floor_div(45.0, r2) == mpfr('14.0')
    assert ctx.floor_div(45.0, r2) == mpfr('14.0')
    assert ctx.floor_div(45.0, 3.1) == mpfr('14.0')
    assert ctx.floor_div(45.0, 4) == mpfr('11.0')
    assert ctx.floor_div(45.0, b) == mpfr('7.0')
    assert ctx.floor_div(45.0, q2) == mpfr('30.0')
    assert ctx.floor_div(45.0, pyq2) == mpfr('30.0')
    assert ctx.floor_div(45.0, 0) == mpfr('inf')
    assert ctx.floor_div(45.0, mpz(0)) == mpfr('inf')
    assert ctx.floor_div(45, r2) == mpfr('14.0')

    pytest.raises(TypeError, lambda: ctx.floor_div(r, 'not'))

    assert r // b == mpfr('7.0')
    assert r // q2 == mpfr('30.0')
    assert r // r2 == mpfr('14.0')

    pytest.raises(TypeError, lambda: r // c2)
    pytest.raises(TypeError, lambda: r // 'not')

    a = mpfr("12.34")
    b = mpfr("45.67")

    assert a//b == mpfr('0.0')
    assert b//a == mpfr('3.0')


def test_mpfr_truediv():
    a = mpfr("12.34")
    b = mpfr("45.67")

    assert a/b == mpfr('0.27019925552879348')
    assert gmpy2.div(a, b) == mpfr('0.27019925552879348')
    assert get_context().div(a, b) == mpfr('0.27019925552879348')
    assert b/a == mpfr('3.7009724473257699')
    assert a/b == 12.34/45.67
    assert a/0 == mpfr('inf')
    assert a/(-0.0) == mpfr('-inf')
    assert a/b == 12.34/45.67
    assert mpfr(10) / z == mpfr('5.0')
    assert mpfr(10) / q == mpfr('6.666666666666667')
    assert mpfr(10) / r == mpfr('6.666666666666667')
    assert b / mpc(4, 4) == mpc('5.7087500000000002-5.7087500000000002j')
    assert get_context().div(b, mpc(4, 4)) == mpc('5.7087500000000002-5.7087500000000002j')

    ans = b / mpc(0, 0)

    assert ans.real == mpfr('inf') and is_nan(ans.imag)

    ans = get_context().div(b, mpc(0, 0))

    assert ans.real == mpfr('inf') and is_nan(ans.imag)

    pytest.raises(TypeError, lambda: b / 'str')
    pytest.raises(TypeError, lambda: get_context().div(b, 'str'))

    assert a / float('Inf') == mpfr('0.0')
    assert -a / float('Inf') == mpfr('-0.0')
    assert float('Inf') / a == mpfr('inf')
    assert float('Inf') / -a == mpfr('-inf')
    assert a / float('-Inf') == mpfr('-0.0')
    assert -a / float('-Inf') == mpfr('0.0')
    assert float('-Inf') / a == mpfr('-inf')
    assert float('-Inf') / -a == mpfr('inf')
    assert is_nan(a / float('nan'))
    assert is_nan(float('nan') / a)
    assert is_nan(float('nan') / mpfr(0))
    assert is_nan(float('nan') / mpfr(0))

    assert a / mpfr('Inf') == mpfr('0.0')
    assert -a / mpfr('Inf') == mpfr('-0.0')
    assert mpfr('Inf') / a == mpfr('inf')
    assert mpfr('Inf') / -a == mpfr('-inf')
    assert a / mpfr('-Inf') == mpfr('-0.0')
    assert -a / mpfr('-Inf') == mpfr('0.0')
    assert mpfr('-Inf') / a == mpfr('-inf')
    assert mpfr('-Inf') / -a == mpfr('inf')
    assert is_nan(a / mpfr('nan'))
    assert is_nan(mpfr('nan') / a)
    assert is_nan(mpfr('nan') / mpfr(0))
    assert is_nan(mpfr('nan') / mpfr(0))


def test_mpfr_mod():
    r = mpfr('0.0') % mpfr('-1.0')
    assert r.is_zero() and r.is_signed()

    a = mpfr("12.34")
    b = mpfr("45.67")

    assert a % 1 == mpfr('0.33999999999999986')
    assert 12.34 % 1 == 0.33999999999999986
    assert a % z == mpfr('0.33999999999999986')
    assert is_nan(b % 0)
    assert is_nan(b % mpz(0))
    assert is_nan(get_context().mod(b, 0))
    assert is_nan(get_context().mod(b, mpz(0)))
    assert is_nan(b % mpfr(0))
    assert is_nan(get_context().mod(b, mpfr(0)))

    get_context().trap_divzero = True

    pytest.raises(gmpy2.DivisionByZeroError, lambda: b % 0)
    pytest.raises(gmpy2.DivisionByZeroError, lambda: b % mpz(0))
    pytest.raises(gmpy2.DivisionByZeroError, lambda: get_context().mod(b, 0))
    pytest.raises(gmpy2.DivisionByZeroError, lambda: get_context().mod(b, mpz(0)))
    pytest.raises(gmpy2.DivisionByZeroError, lambda: b % mpfr(0))
    pytest.raises(gmpy2.DivisionByZeroError, lambda: get_context().mod(b, mpfr(0)))

    get_context().trap_divzero = False

    pytest.raises(TypeError, lambda: b % 'str')
    pytest.raises(TypeError, lambda: get_context().mod(b,'str'))
    pytest.raises(TypeError, lambda: b % mpc(3, 5))

    assert is_nan(mpfr('nan') % a)
    assert is_nan(a % mpfr('nan'))
    assert is_nan(mpfr('inf') % a)
    assert a % mpfr('inf') == mpfr('12.34')

    get_context().trap_invalid = True

    pytest.raises(gmpy2.InvalidOperationError, lambda: mpfr('nan') % a)
    pytest.raises(gmpy2.InvalidOperationError, lambda: a % mpfr('nan'))
    pytest.raises(gmpy2.InvalidOperationError, lambda: mpfr('inf') % a)
    pytest.raises(gmpy2.InvalidOperationError, lambda: a % mpfr('inf'))

    get_context().trap_invalid = False

    pytest.raises(TypeError, lambda: get_context().mod(a, b, 5))

    assert get_context().mod(a, -mpfr('inf')) == mpfr('-inf')


def test_mpfr_pow():
    r1, r2 = mpfr(5.0), mpfr(2.5)
    ctx = gmpy2.get_context()

    assert r1 ** mpz(2) == mpfr('25.0')
    assert r2 ** mpz(2) == mpfr('6.25')
    assert r2 ** 2 == mpfr('6.25')
    assert pow(r1, r2) == mpfr('55.901699437494742')
    assert ctx.pow(r1, r2) == mpfr('55.901699437494742')
    assert ctx.pow(r1, r2) == r1 ** r2

    pytest.raises(TypeError, lambda: pow(r1, r2, 5))
    pytest.raises(TypeError, lambda: ctx.pow(r1, r2, 5))

    assert pow(r1, 4) == mpfr('625.0')
    assert ctx.pow(r1, 4) == mpfr('625.0')


def test_mpfr_subnormalize():
    gmpy2.set_context(gmpy2.ieee(64))

    zeroes = '0.' + '0' * 323

    assert mpfr(zeroes + '2470328229206232720') == mpfr('0.0')
    assert mpfr(zeroes + '2470328229206232721') == mpfr('4.9406564584124654e-324')
    assert mpfr(zeroes + '247032822920623272088') == mpfr('0.0')
    assert mpfr(zeroes + '247032822920623272089') == mpfr('4.9406564584124654e-324')

    gmpy2.set_context(gmpy2.ieee(32))

    def fmt(v):
        return '{:.23b}'.format(v)


    a = mpfr('0x1p-126')

    assert fmt(a) == '1.00000000000000000000000p-126'
    assert fmt(gmpy2.next_below(a)) == '1.11111111111111111111110p-127'
    assert fmt(gmpy2.next_above(a)) == '1.00000000000000000000001p-126'
    assert fmt(gmpy2.next_below(0)) == '-1.00000000000000000000000p-149'
    assert fmt(gmpy2.next_above(0)) == '1.00000000000000000000000p-149'
    assert fmt(gmpy2.next_toward(a, -10)) == '1.11111111111111111111110p-127'
    assert fmt(gmpy2.next_toward(a, 10)) == '1.00000000000000000000001p-126'

    gmpy2.set_context(gmpy2.context())


def test_mpfr_as_integer_ratio():
    assert mpfr('1.1e+2').as_integer_ratio() == (mpz(110), mpz(1))

    pytest.raises(ValueError, lambda: mpfr('nan').as_integer_ratio())
    pytest.raises(OverflowError, lambda: mpfr('inf').as_integer_ratio())


def test_mpfr_round():
    pytest.raises(TypeError, lambda: round(mpfr('1.0'), "spam"))

    r = round(mpfr('-0.0'), 123)
    assert r.is_zero() and r.is_signed()

    assert round(mpfr('12.34'), -1) == mpfr('10.0')

    r = mpfr(4.55)

    assert round(r, 1) == mpfr('4.5')
    assert round(mpfr('5.1')) == mpz(5)

    pytest.raises(ValueError, lambda: round(nan()))
    pytest.raises(OverflowError, lambda: round(inf()))


def test_mpfr_as_mantissa_exp():
    r = mpfr(4.55)

    assert r.as_mantissa_exp() == (mpz(5122844576133939), mpz(-50))
    assert mpfr(0).as_mantissa_exp() == (mpz(0), mpz(1))

    pytest.raises(OverflowError, lambda: inf().as_mantissa_exp())
    pytest.raises(ValueError, lambda: nan().as_mantissa_exp())


def test_mpfr_as_simple_fraction():
    r = mpfr(4.55)

    assert r.as_simple_fraction() == mpq(91,20)

    pytest.raises(ValueError, lambda: r.as_simple_fraction(precision=100))
    pytest.raises(TypeError, lambda: r.as_simple_fraction(nosense=10))


def test_mpfr_real_imag():
    r = mpfr(4.55)
    a = mpfr('12.34')

    assert r.imag == mpfr('0.0')
    assert r.real == mpfr('4.5499999999999998')
    assert a.real == mpfr('12.34')
    assert a.imag == mpfr('0.0')


def test_mpfr_conjugate():
    r = mpfr(4.55)
    a = mpfr('12.34')

    assert r.conjugate() == r
    assert a.conjugate() == a


def test_mpfr_pickle():
    a = mpfr('12.34')

    assert pickle.loads(pickle.dumps(a)) == a
    assert pickle.loads(pickle.dumps(mpfr("inf"))) == mpfr('inf')
    assert pickle.loads(pickle.dumps(mpfr("-inf"))) == mpfr('-inf')
    assert is_nan(pickle.loads(pickle.dumps(mpfr("nan"))))
    assert pickle.loads(pickle.dumps(mpfr(0))) == mpfr('0.0')



from fractions import Fraction

import pytest
from hypothesis import example, given, settings
from hypothesis.strategies import complex_numbers
from supportclasses import a, b, c, cx, d, q, r, z

import gmpy2
from gmpy2 import (cmp, cmp_abs, from_binary, get_context, is_nan, mpc,
                   mpc_random, mpfr, mpq, mpz, nan, random_state, to_binary)


def test_mpc_cmp():
    pytest.raises(TypeError, lambda: cmp(mpc(1,2), mpc(3,4)))
    assert cmp_abs(mpc(1,2), mpc(3,4)) == -1
    assert cmp_abs(mpc(1,2), mpc(1,2)) == 0
    assert cmp_abs(mpc(3,4), mpc(1,2)) == 1
    gmpy2.get_context().clear_flags()
    assert gmpy2.get_context().erange is False
    assert cmp_abs(mpc(nan(),1), mpc(4.5)) == 0
    assert gmpy2.get_context().erange is True


def test_mpc_comparisons():
    a = mpz(123)
    c = mpc(4.5,6)

    assert (c == a, c != a) == (False, True)

    pytest.raises(TypeError, lambda: c < a)

    cnan = mpc(mpfr('nan'), 6)

    assert (c == cnan, c != cnan) == (False, True)

    pytest.raises(TypeError, lambda: c > cnan)

    cx = complex(4.5,6)

    assert (c == cx, c != cx) == (True, False)

    pytest.raises(TypeError, lambda: c > cx)

    from supportclasses import c

    assert (c == cx) is False


def test_mpc_conversion():
    x = mpc(a)
    assert isinstance(x, mpc)
    assert x == 42+67j

    pytest.raises(TypeError, lambda: mpc(b))
    pytest.raises(TypeError, lambda: mpc(c))
    pytest.raises(TypeError, lambda: mpc(d))

    assert mpc(mpfr(5.6), precision=(0,0)) == mpc('5.5999999999999996+0.0j')
    assert mpc(Fraction(4,5)) == mpc('0.80000000000000004+0.0j')
    assert mpc(b'5+6j') == mpc('5.0+6.0j')
    assert mpc('5+6j') == mpc('5.0+6.0j')

    pytest.raises(ValueError, lambda: mpc('\xc3'))
    pytest.raises(ValueError, lambda: mpc('notanumber'))
    pytest.raises(ValueError, lambda: mpc('\0'))

    assert mpc('(5+6j)') == mpc('5.0+6.0j')
    assert mpc('   5+6j   ') == mpc('5.0+6.0j')

    pytest.raises(ValueError, lambda: mpc('5+6ji'))

    assert str(mpc(5,6)) == '5.0+6.0j'
    assert complex(mpc(4,5)) == (4+5j)
    assert complex(mpc(4,5)) == (4+5j)


def test_mpc_creation():
    ctx = gmpy2.get_context()
    ctx.clear_flags()
    a = mpc("1.2")
    assert a.rc == (-1, 0)
    assert ctx.inexact
    ctx.clear_flags()
    a = mpc("(1 2)")
    assert a == 1 + 2j
    assert a.rc == (0, 0)
    assert ctx.inexact is False
    ctx.clear_flags()
    a = mpc("1   + 2.1  j")
    assert a == 1 + 2.1j
    assert a.rc == (0, 1)
    assert ctx.inexact

    c = mpc('67+87j', precision=70)

    assert c.precision == (70, 70)
    assert c.real.precision == 70
    assert c.imag.precision == 70

    c = mpc('42e56+42.909j', precision=(45,300))

    assert c.precision == (45, 300)
    assert c.real.precision == 45
    assert c.imag.precision == 300

    x = mpc("1.3142123+4.3e-1001j", precision=(70,37))

    assert mpc(x.real, x.imag, precision=(70,37)) == x

    assert mpc('1_2+4_5j') == mpc('12.0+45.0j')


def test_mpc_random():
    assert (mpc_random(random_state(42))
            == mpc('0.86555158787663011+0.4422082613292212j'))


def test_mpc_to_from_binary():
    x = mpc("0+0j")
    assert x == from_binary(to_binary(x))
    x = mpc("1+0j")
    assert x == from_binary(to_binary(x))
    x = mpc("-1+0j")
    assert x == from_binary(to_binary(x))
    x = mpc("0+1j")
    assert x == from_binary(to_binary(x))
    x = mpc("0-1j")
    assert x == from_binary(to_binary(x))
    x = mpc("inf+0j")
    assert x == from_binary(to_binary(x))
    x = mpc("0+infj")
    assert x == from_binary(to_binary(x))
    x = mpc("inf-infj")
    assert x == from_binary(to_binary(x))
    x = mpc("inf+nanj")
    y = from_binary(to_binary(x))
    assert x.real == y.real
    assert is_nan(y.imag)
    x = mpc("-inf+0j")
    assert x == from_binary(to_binary(x))
    x = mpc("0-infj")
    assert x == from_binary(to_binary(x))
    x = mpc("-inf-infj")
    assert x == from_binary(to_binary(x))
    x = mpc("-inf+nanj")
    y = from_binary(to_binary(x))
    assert x.real == y.real
    assert is_nan(y.imag)
    x = mpc("nan+0j")
    y = from_binary(to_binary(x))
    assert x.imag == y.imag
    assert is_nan(y.real)
    x = mpc("0+nanj")
    y = from_binary(to_binary(x))
    assert x.real == y.real
    assert is_nan(y.imag)
    x = mpc("nan-infj")
    y = from_binary(to_binary(x))
    assert x.imag == y.imag
    assert is_nan(y.real)
    x = mpc("nan+nanj")
    y = from_binary(to_binary(x))
    assert is_nan(y.real)
    assert is_nan(y.imag)
    get_context().real_prec=100
    get_context().imag_prec=110
    assert (from_binary(to_binary(mpc("1.3-4.7j"))) ==
            mpc('1.2999999999999999999999999999994-4.7000000000000000000000000000000025j',
                (100,110)))


def test_mpc_format():
    gmpy2.set_context(gmpy2.context())

    c, c1 = mpc(mpq(1/3), 5), mpc(-1, -2)

    assert '{:>20}'.format(c) == '  0.333333+5.000000j'
    assert '{:<20}'.format(c) == '0.333333+5.000000j  '
    assert '{:^20}'.format(c) == ' 0.333333+5.000000j '

    pytest.raises(ValueError, lambda: '{:<<20}'.format(c))

    assert '{:>+20}'.format(c) == ' +0.333333+5.000000j'

    pytest.raises(ValueError, lambda: '{:+^20}'.format(c))

    assert '{:.10f}'.format(c) == '0.3333333333+5.0000000000j'

    pytest.raises(ValueError, lambda: '{:Z.10f}'.format(c))
    pytest.raises(ValueError, lambda: '{:Z 10}'.format(c))

    assert '{:Z}'.format(c) == '0.333333+5.000000j'
    assert '{:U}'.format(c) == '0.333334+5.000000j'

    pytest.raises(ValueError, lambda: '{:PU}'.format(c))

    assert '{:UP}'.format(c) == '0.333334+5.000000j'

    pytest.raises(ValueError, lambda: '{:PP}'.format(c))

    assert '{:G}'.format(c) == '0.333333+5.0j'
    assert '{:M}'.format(c) == '(0.333333 5.000000)'
    assert '{:b}'.format(c) == '1.0101010101010101010101010101010101010101010101010101p-2+1.01p+2j'
    assert '{:a}'.format(c) == '0x5.5555555555554p-4+0x5p+0j'
    assert '{:e}'.format(c) in ('3.3333333333333331e-01+5e+00j', '3.3333333333333331e-01+5.0000000000000000e+00j')
    assert '{:M}'.format(c1) == '(-1.000000 -2.000000)'


def test_mpc_repr():
    c = mpc('1.2999999999999999999999999999994-4.7000000000000000000000000000000025j',(100,110))
    assert repr(c) == "mpc('1.2999999999999999999999999999994-4.7000000000000000000000000000000025j',(100,110))"
    assert repr(mpc("1+nanj")) == "mpc('1.0+nanj')"
    assert repr(mpc("infj")) == "mpc('0.0+infj')"


def test_mpc_digits():
    c = mpc(mpq(1/3), 5)

    assert c.digits(8) == (('2525252525252525250', 0, 53), ('5000000000000000000', 1, 53))
    assert c.digits(8, 2) == (('25', 0, 53), ('50', 1, 53))

    pytest.raises(ValueError, lambda: c.digits(8, -2))
    pytest.raises(ValueError, lambda: c.digits(0))


def test_mpc_abs():
    assert abs(mpc(-1,0)) == mpfr('1.0')
    assert abs(-1+0j) == 1.0
    assert abs(mpc(1,1)) == mpfr('1.4142135623730951')

    ctx = gmpy2.get_context()
    ctx.clear_flags()

    c = mpc('nan+0.0j')

    assert is_nan(c.real) and c.imag == 0.0
    assert ctx.invalid

    ctx.clear_flags()

    c = mpc('nan+0j')

    assert is_nan(c.real) and c.imag == 0.0
    assert ctx.invalid

    ctx.clear_flags()

    assert is_nan(abs(mpc('nanj'))) and ctx.invalid

    ctx.clear_flags()

    assert abs(mpc('inf+10j')) == mpfr('inf')
    assert abs(mpc('-infj')) == mpfr('inf')

    a = mpc('nan+infj')

    assert abs(a) and not ctx.invalid

    a = mpc('-inf+nanj')

    assert abs(a) and not ctx.invalid


def test_mpc_add():
    a = mpz(123)
    aj = mpc(1+2j)
    bj = mpc(4+5j)

    assert aj + bj == mpc('5.0+7.0j')
    assert bj + aj == mpc('5.0+7.0j')
    assert aj + a == mpc('124.0+2.0j')
    assert a + aj == mpc('124.0+2.0j')
    assert aj + 1 == mpc('2.0+2.0j')
    assert 1 + aj == mpc('2.0+2.0j')
    assert aj + 0 == mpc('1.0+2.0j')
    assert 0 + aj == mpc('1.0+2.0j')
    assert -1 + aj == mpc('0.0+2.0j')
    assert aj - 1 == mpc('0.0+2.0j')
    assert aj + 1.2 == 2.2 + 2j

    assert aj + float('inf') == mpc('inf+2.0j')
    assert aj + float('-inf') == mpc('-inf+2.0j')

    x = aj + float('nan')

    assert is_nan(x.real) and x.imag == 2.0


def test_mpc_sub():
    pytest.raises(TypeError, lambda: mpc(1,2) - 'a')

    assert mpfr(1) - mpc(1,2) == mpc('0.0-2.0j')
    assert mpc(1,2) - mpfr(1) == mpc('0.0+2.0j')
    assert mpc(1,2) - 1+0j == mpc('0.0+2.0j')
    assert 1+0j - mpc(1,2) == mpc('0.0-2.0j')
    assert mpc(1,2) - z == mpc('-1.0+2.0j')
    assert mpc(1,2) - q == mpc('-0.5+2.0j')
    assert mpc(1,2) - r == mpc('-0.5+2.0j')
    assert mpc(1,2) - cx == mpc('-41.0-65.0j')

    a = mpz(123)
    aj = mpc(1+2j)
    bj = mpc(4+5j)

    assert aj - bj == mpc('-3.0-3.0j')
    assert bj - aj == mpc('3.0+3.0j')
    assert aj - a == mpc('-122.0+2.0j')
    assert a - aj == mpc('122.0-2.0j')
    assert aj - 1 == mpc('0.0+2.0j')
    assert 1 - aj == mpc('0.0-2.0j')
    assert 0 - aj == mpc('-1.0-2.0j')
    assert aj - 0 == mpc('1.0+2.0j')
    assert aj - -1 == mpc('2.0+2.0j')
    assert -1 - aj == mpc('-2.0-2.0j')
    assert aj - 1.2 == (1+2j) - 1.2

    assert float('inf') - aj == mpc('inf-2.0j')
    assert aj - float('inf') == mpc('-inf+2.0j')
    assert aj - float('-inf') == mpc('inf+2.0j')

    x = aj - float('nan')

    assert is_nan(x.real) and x.imag == 2.0


def test_mpc_mul():
    pytest.raises(TypeError, lambda: mpc(1,2) * 'a')

    assert mpfr(1) * mpc(1,2) == mpc('1.0+2.0j')
    assert mpc(1,2) * mpfr(1) == mpc('1.0+2.0j')
    assert mpc(1,2) * mpfr(-1) == mpc('-1.0-2.0j')
    assert mpc(1,2) * (1+0j) == mpc('1.0+2.0j')
    assert (1+0j) * mpc(1,2) == mpc('1.0+2.0j')
    assert mpc(1,2) * z == mpc('2.0+4.0j')
    assert mpc(1,2) * q == mpc('1.5+3.0j')
    assert mpc(1,2) * r == mpc('1.5+3.0j')
    assert mpc(1,2) * cx == mpc('-92.0+151.0j')

    a = mpz(123)
    aj = mpc(1+2j)
    bj = mpc(4+5j)

    assert aj * bj == mpc('-6.0+13.0j')
    assert bj * aj == mpc('-6.0+13.0j')
    assert aj * a == mpc('123.0+246.0j')
    assert a * aj == mpc('123.0+246.0j')
    assert aj * -1 == mpc('-1.0-2.0j')
    assert aj * (0.0+1j) == mpc('-2.0+1.0j')

    assert aj * float('inf') == mpc('inf+infj')
    assert aj * float('-inf') == mpc('-inf-infj')

    for x in [aj * float('nan'), mpc(0,0) * float('inf'),
              mpc(0,0) * float('-inf'), mpc(0,0) * float('nan')]:
        assert all(is_nan(_) for _ in [x.real, x.imag])


def test_mpc_divmod():
    pytest.raises(TypeError, lambda: divmod(mpc(1),'a'))

    ctx = gmpy2.context()

    pytest.raises(TypeError, lambda: ctx.divmod(mpc(1,2),mpc(3,4)))
    pytest.raises(TypeError, lambda: divmod(mpc(1,2), mpc(1,2)))
    pytest.raises(TypeError, lambda: ctx.divmod(mpc(1,2),mpc(3,4)))

    aj = mpc(1+2j)
    bj = mpc(4+5j)

    with pytest.raises(TypeError):
        divmod(aj, bj)


def test_mpc_div():
    a = mpz(123)
    aj = mpc(1+2j)
    bj = mpc(4+5j)

    assert aj / bj == mpc('0.34146341463414637+0.073170731707317069j')
    assert gmpy2.div(aj, bj) == mpc('0.34146341463414637+0.073170731707317069j')

    with pytest.raises(TypeError):
        aj // bj

    assert aj / a == mpc('0.008130081300813009+0.016260162601626018j')
    assert a / aj == mpc('24.600000000000001-49.200000000000003j')
    assert aj / 0 == mpc('inf+infj')
    assert mpc('2.0+2.0j') / z == mpc('1.0+1.0j')
    assert mpc('2.0+2.0j') / q == mpc('1.3333333333333333+1.3333333333333333j')
    assert mpc('2.0+2.0j') / r == mpc('1.3333333333333333+1.3333333333333333j')
    assert mpc(15,15) / cx == mpc('0.26147449224372299-0.059971213817367662j')

    with gmpy2.context(trap_divzero=True):
        with pytest.raises(gmpy2.DivisionByZeroError):
            mpc(15, 15)/mpc(0)

    assert aj / float('inf') == mpc('0.0+0.0j')
    assert aj / float('-inf') == mpc('-0.0-0.0j')
    assert float('inf') / aj == mpc('inf-infj')
    assert float('-inf') / aj == mpc('-inf+infj')


def test_mpc_mod():
    a = mpz(123)
    aj = mpc(1+2j)
    bj = mpc(4+5j)

    with pytest.raises(TypeError):
        aj % bj
    with pytest.raises(TypeError):
        aj % a
    with pytest.raises(TypeError):
        a % aj


def test_mpc_pow():
    c1, c2 = mpc(2,5), mpc(5,2)
    ctx = gmpy2.get_context()

    assert ctx.pow(complex(2,5), complex(5,2)) == mpc('-416.55882051164394+44.334999625388825j')
    assert pow(c1, c2) == mpc('-416.55882051164394+44.334999625388825j')
    assert ctx.pow(c1, c2) == mpc('-416.55882051164394+44.334999625388825j')
    assert ctx.pow(c1, c2) == c1 ** c2

    pytest.raises(TypeError, lambda: pow(c1, c2, 5))

    assert pow(c1, 5) == mpc('4282.0-1475.0j')
    assert c1 ** mpz(5) == mpc('4282.0-1475.0j')
    assert c1 ** mpfr(2.5) == mpc('-66.373652915897722+11.111336616269842j')


@settings(max_examples=1000)
@given(complex_numbers(allow_nan=False))
@example(complex())
@example(complex(-1))
@example(complex(-2))
def test_mpc_hash(c):
    assert hash(mpc(c)) == hash(c)


def test_mpc_exc():
    gmpy2.set_context(gmpy2.ieee(32))

    ctx = gmpy2.get_context()
    ctx.trap_overflow = True
    ctx.trap_underflow = True

    c = mpc(0.1 + 0.1j)

    pytest.raises(gmpy2.UnderflowResultError, lambda: c**201)
    pytest.raises(gmpy2.OverflowResultError, lambda: c**-201)

    ctx.trap_inexact = True

    pytest.raises(gmpy2.InexactResultError, lambda: mpc(0.25)**0.25)

    ctx.trap_invalid = True

    pytest.raises(gmpy2.InvalidOperationError, lambda: mpc(mpfr('nan')))



import math
import numbers
import pickle
from fractions import Fraction

from hypothesis import assume, example, given, settings
from hypothesis.strategies import booleans, integers, sampled_from
from pytest import mark, raises
from supportclasses import a, b, c, d, q, z

import gmpy2
from gmpy2 import (cmp, cmp_abs, from_binary, is_nan, is_prime, mp_version,
                   mpc, mpfr, mpq, mpz, mpz_random, mpz_rrandomb, mpz_urandomb,
                   next_prime, pack, random_state, to_binary, unpack, xmpz)


def test_mpz_to_bytes_interface():
    x = mpz(1)
    with raises(TypeError):
        x.to_bytes(1, 2, 3)
    with raises(TypeError):
        x.to_bytes(1, 2)
    with raises(TypeError):
        x.to_bytes('spam')
    with raises(TypeError):
        x.to_bytes(a=1, b=2, c=3, d=4)
    with raises(TypeError):
        x.to_bytes(2, length=2)
    with raises(TypeError):
        x.to_bytes(2, 'big', byteorder='big')
    with raises(TypeError):
        x.to_bytes(spam=1)

    with raises(ValueError):
        x.to_bytes(2, 'spam')
    with raises(ValueError):
        x.to_bytes(-1)

    assert x.to_bytes(2) == x.to_bytes(length=2)
    assert x.to_bytes(2, byteorder='little') == x.to_bytes(2, 'little')

    assert x.to_bytes() == x.to_bytes(1)
    assert x.to_bytes() == x.to_bytes(1, 'big')
    assert x.to_bytes() == x.to_bytes(signed=False)


@settings(max_examples=1000)
@given(integers(), integers(min_value=0, max_value=10000),
       sampled_from(['big', 'little']), booleans())
@example(0, 0, 'big', False)
@example(0, 0, 'little', False)
@example(0, 1, 'big', False)
@example(128, 1, 'big', True)
@example(128, 1, 'little', True)
@example(-129, 1, 'big', True)
@example(-129, 1, 'little', True)
@example(-1, 0, 'big', True)
@example(-2, 0, 'big', True)
@example(-2, 0, 'little', True)
@example(42, 1, 'big', False)
@example(42, 1, 'little', False)
@example(42, 3, 'big', False)
@example(42, 3, 'little', False)
@example(1000, 2, 'big', False)
@example(1000, 4, 'big', False)
@example(-2049, 1, 'big', True)
@example(-65281, 3, 'big', True)
@example(-65281, 3, 'little', True)
def test_mpz_to_bytes(x, length, byteorder, signed):
    try:
        rx = x.to_bytes(length, byteorder, signed=signed)
    except OverflowError:
        with raises(OverflowError):
            mpz(x).to_bytes(length, byteorder, signed=signed)
    else:
        assert rx == mpz(x).to_bytes(length, byteorder, signed=signed)


def test_mpz_from_bytes_interface():
    with raises(TypeError):
        mpz.from_bytes()
    with raises(TypeError):
        mpz.from_bytes(1, 2, 3)
    with raises(TypeError):
        mpz.from_bytes(b'', 2)
    with raises(TypeError):
        mpz.from_bytes(1)
    with raises(TypeError):
        mpz.from_bytes(b'', bytes=b'')
    with raises(TypeError):
        mpz.from_bytes(b'', 'big', byteorder='big')
    with raises(TypeError):
        mpz.from_bytes(b'', spam=1)
    with raises(TypeError):
        mpz.from_bytes(a=1, b=2, c=3, d=4)

    with raises(ValueError):
        mpz.from_bytes(b'', 'spam')

    assert mpz.from_bytes(b'\x01', byteorder='little') == mpz.from_bytes(b'\x01', 'little')

    assert mpz.from_bytes(b'\x01') == mpz.from_bytes(bytes=b'\x01')
    assert mpz.from_bytes(b'\x01') == mpz.from_bytes(b'\x01', 'big')
    assert mpz.from_bytes(b'\x01') == mpz.from_bytes(b'\x01', signed=False)


@settings(max_examples=1000)
@given(integers(), integers(min_value=0, max_value=10000),
       sampled_from(['big', 'little']), booleans())
@example(0, 0, 'big', False)
@example(0, 0, 'little', False)
@example(0, 1, 'big', False)
@example(128, 1, 'big', True)
@example(128, 1, 'little', True)
@example(-129, 1, 'big', True)
@example(-129, 1, 'little', True)
@example(-1, 0, 'big', True)
@example(-1, 1, 'big', True)
@example(-1, 1, 'little', True)
@example(-2, 0, 'big', True)
@example(-2, 0, 'little', True)
@example(-1, 3, 'big', True)
@example(-2, 3, 'big', True)
@example(-2, 5, 'little', True)
def test_mpz_from_bytes(x, length, byteorder, signed):
    try:
        bytes = x.to_bytes(length, byteorder, signed=signed)
    except OverflowError:
        assume(False)
    else:
        rx = int.from_bytes(bytes, byteorder, signed=signed)
        assert rx == mpz.from_bytes(bytes, byteorder, signed=signed)
        assert rx == mpz.from_bytes(bytearray(bytes), byteorder, signed=signed)
        assert rx == mpz.from_bytes(list(bytes), byteorder, signed=signed)


def test_mpz_as_integer_ratio():
    assert mpz(3).as_integer_ratio() == (mpz(3), mpz(1))


def test_mpz_numbers_abc():
    assert isinstance(mpz(2), numbers.Integral)


def test_mpz_pickling():
    for proto in range(pickle.HIGHEST_PROTOCOL + 1):
        for x in [mpz(12346789), mpz(-12346789), mpz(0)]:
            assert pickle.loads(pickle.dumps(x, protocol=proto)) == x


@settings(max_examples=1000)
@given(integers(), integers())
@example(0, 0)
@example(0, 1)
@example(-11, 75)
@example(14, 105)
@example(64, 123456789012345678901234567890)
@example(321, -123)
def test_mpz_arithmetics(i, z):
    i, z = map(mpz, [i, z])

    assert int(i) + int(z) == i + z
    assert int(z) + int(i) == z + i

    assert int(i) - int(z) == i - z
    assert int(z) - int(i) == z - i

    assert int(i) * int(z) == i * z
    assert int(z) * int(i) == z * i

    # Test all permutations of floor division
    if z:
        assert int(i) // int(z) == i // z
        assert int(i) % int(z) == i % z
        assert i % int(z) == i % z
        assert int(i) % z == i % z
        assert divmod(int(i), int(z)) == divmod(i, z)
        assert divmod(i, int(z)) == divmod(i, z)
        assert divmod(int(i), z) == divmod(i, z)

    if i:
        assert int(z) // int(i) == z // i
        assert int(z) % int(i) == z % i
        assert z % int(i) == z % i
        assert int(z) % i == z % i
        assert divmod(int(z), int(i)) == divmod(z, i)
        assert divmod(z, int(i)) == divmod(z, i)
        assert divmod(int(z), i) == divmod(z, i)


@settings(max_examples=1000)
@given(integers(min_value=0),
       integers(min_value=1, max_value=100000))
def test_mpz_pack_unpack_bulk(x, n):
    lst = unpack(x, n)
    assert pack(lst, n) == x


def test_mpz_pack_unpack():
    x = mpz(0)
    assert all((x == pack(unpack(x,i),i) for i in range(1,100)))
    x = mpz(1)
    assert all((x == pack(unpack(x,i),i) for i in range(1,100)))
    x = mpz(2)
    assert all((x == pack(unpack(x,i),i) for i in range(1,100)))
    x = mpz(3141592635)
    assert all((x == pack(unpack(x,i),i) for i in range(1,100)))
    x = mpz(1234567891234567890000000000000000000000000000000000000123)
    assert all((x == pack(unpack(x,i),i) for i in range(1,100)))
    x = mpz(1) << 500
    assert all((x == pack(unpack(x,i),i) for i in range(1,200)))
    x -= 1
    assert all((x == pack(unpack(x,i),i) for i in range(1,200)))

    raises(TypeError, lambda: pack(x))
    raises(TypeError, lambda: pack(1, 1))
    raises(TypeError, lambda: pack([mpz(1), mpz(-1)], 2))

    raises(TypeError, lambda: unpack(x))
    raises(TypeError, lambda: unpack([], 1))
    raises(ValueError, lambda: unpack(-1, 1))


def test_mpz_cmp():
    assert cmp(0, mpz(0)) == 0
    assert cmp(1, mpz(0)) == 1
    assert cmp(0, mpz(1)) == -1
    assert cmp(-1, mpz(0)) == -1
    assert cmp(0, mpz(-1)) == 1

    assert cmp_abs(mpz(0), 0) == 0
    assert cmp_abs(mpz(1), 0) == 1
    assert cmp_abs(mpz(0), 1) == -1
    assert cmp_abs(mpz(-1), 0) == 1
    assert cmp_abs(mpz(0), -1) == -1

    a = mpz(-10)
    assert cmp_abs(a, 0) == 1
    assert a == mpz(-10)
    assert cmp_abs(100, a) == 1
    assert a == mpz(-10)

    assert cmp(mpz(2), z) == 0
    assert cmp(z, mpz(3)) == -1
    assert cmp(mpz(1), q) == -1
    assert cmp(mpz(1), mpz(q)) == 0


def test_mpz_comparisons():
    from supportclasses import q

    assert mpz(2) == z
    assert (z == mpz(3)) is False
    assert (mpz(1) == q) is False
    assert mpz(1) == mpz(q)

    a = mpz(123)
    b = mpz(456)
    c = mpz(a)
    q = mpq(4, 5)

    assert a == mpz(123)
    assert b == mpz(456)
    assert c is a
    assert c==a
    assert (c>a) is False
    assert (c<a) is False
    assert (a>b) is False
    assert a<b
    assert not mpz(0)
    assert not a is False
    assert (mpz(1) == None) is False
    assert (mpz(1) == '1') is False
    assert (mpz(1) == 'abc') is False
    assert [mpz(23), None].count(None) == 1
    assert (a == q, a != q, a > q, a >= q, a < q, a <= q) == (False, True, True, True, False, False)

    q = mpq(123, 1)

    assert (a == q, a != q, a > q, a >= q, a < q, a <= q) == (True, False, False, True, False, True)

    gmpy2.context().trap_divzero == False
    f = float('inf')

    assert (a == f, a != f, a > f, a >= f, a < f, a <= f) == (False, True, False, False, True, True)
    assert (f == a, f != a, f > a, f >= a, f < a, f <= a) == (False, True, True, True, False, False)

    f = float('-inf')

    assert (a == f, a != f, a > f, a >= f, a < f, a <= f) == (False, True, True, True, False, False)
    assert (f == a, f != a, f > a, f >= a, f < a, f <= a) == (False, True, False, False, True, True)

    f = float('nan')

    assert (a == f, a != f, a > f, a >= f, a < f, a <= f) == (False, True, False, False, False, False)
    assert (f == a, f != a, f > a, f >= a, f < a, f <= a) == (False, True, False, False, False, False)

    r = mpfr('inf')

    assert (a == r, a != r, a > r, a >= r, a < r, a <= r) == (False, True, False, False, True, True)

    r = mpfr('-inf')

    assert (a == r, a != r, a > r, a >= r, a < r, a <= r) == (False, True, True, True, False, False)

    r = mpfr('nan')

    assert (a == r, a != r, a > r, a >= r, a < r, a <= r) == (False, True, False, False, False, False)


def test_mpz_conversion():
    x = mpz(a)
    assert isinstance(x, mpz)
    assert x == 1
    raises(TypeError, lambda: mpz(b))
    raises(TypeError, lambda: mpz(c))
    raises(TypeError, lambda: mpz(d))

    assert float(mpz(1)) == 1.0
    raises(OverflowError, lambda: float(mpz(99**199)))
    assert mpz(xmpz(1)) == mpz(1)

    assert int(mpz(-3)) == -3

    assert int(mpz(11)) is int(mpz(11))


def test_mpz_create():
    assert mpz() == mpz(0)
    assert mpz(0) == mpz(0)
    assert mpz(1) == mpz(1)
    assert mpz(-1) == mpz(-1)
    assert mpz(2**15-2) == mpz(32766)
    assert mpz(2**15-1) == mpz(32767)
    assert mpz(2**15) == mpz(32768)
    assert mpz(2**15+1) == mpz(32769)
    assert mpz(2**15+2) == mpz(32770)
    assert mpz(2**30-2) == mpz(1073741822)
    assert mpz(2**30-1) == mpz(1073741823)
    assert mpz(2**30) == mpz(1073741824)
    assert mpz(2**30+1) == mpz(1073741825)
    assert mpz(2**30+2) == mpz(1073741826)
    assert mpz(2**16-2) == mpz(65534)
    assert mpz(2**16-1) == mpz(65535)
    assert mpz(2**16) == mpz(65536)
    assert mpz(2**16+1) == mpz(65537)
    assert mpz(2**16+2) == mpz(65538)
    assert mpz(1000000000000) == mpz(1000000000000)
    assert mpz(-1000000000000) == mpz(-1000000000000)

    raises(ValueError, lambda: mpz(''))
    raises(ValueError, lambda: mpz('a'))

    assert mpz('a',16) == mpz(10)

    raises(ValueError, lambda: mpz('z',16))

    assert mpz('0b1101') == mpz(13)
    assert mpz('0b1101',2) == mpz(13)
    assert mpz('1101',2) == mpz(13)
    assert mpz('0b0010') == mpz(2)
    assert mpz('0b0010',2) == mpz(2)

    raises(ValueError, lambda: mpz('0b0b10',2))
    raises(ValueError, lambda: mpz('0b0b10'))
    raises(ValueError, lambda: mpz('0b0012'))

    assert mpz('0o0012') == mpz(10)
    assert mpz('0o0012',8) == mpz(10)
    assert mpz('12',8) == mpz(10)
    assert mpz('0x12') == mpz(18)
    assert mpz('0x12',16) == mpz(18)
    assert mpz('12',16) == mpz(18)
    assert mpz('-1') == mpz(-1)
    assert mpz('+1') == mpz(1)
    assert mpz('  0xA', base=0) == mpz(10)

    raises(ValueError, lambda: mpz(float('nan')))
    raises(OverflowError, lambda: mpz(float('inf')))
    raises(OverflowError, lambda: mpz(float('-inf')))
    raises(TypeError, lambda: mpz(12, base=16))

    assert mpz('12', base=16) == mpz(18)

    raises(ValueError, lambda: mpz('\xff'))
    raises(ValueError, lambda: mpz('\x0cf'))
    raises(ValueError, lambda: mpz('\0xff'))

    assert mpz(b'12') == mpz(12)

    raises(TypeError, lambda: mpz(None))
    raises(TypeError, lambda: mpz(None,base=10))
    raises(ValueError, lambda: mpz('99',base=100))
    raises(TypeError, lambda: mpz('99',base='a'))

    assert mpz('99',base=10) == mpz(99)
    assert mpz(xmpz(5)) == mpz(5)

    raises(ValueError, lambda: mpz('ы'))
    raises(ValueError, lambda: mpz(bytes('ы', encoding='utf-8')))

    assert mpz(3.14) == mpz(3)
    assert mpz(mpq(17,3)) == mpz(5)
    assert mpz(23) == mpz(23)
    assert mpz(-23) == mpz(-23)

    x = 1000*1000*1000*1000*1000*1000*1000

    assert mpz(x) == 1000000000000000000000
    assert mpz(0.0) == mpz(0)
    assert mpz(-0.0) == mpz(0)

    raises(ValueError, lambda: mpz(float("nan")))
    raises(OverflowError, lambda: mpz(float("inf")))
    raises(OverflowError, lambda: mpz(float("-inf")))

    assert mpz("0") == mpz(0)
    assert mpz("-0") == mpz(0)

    raises(ValueError, lambda: mpz("hi"))

    assert mpz("123456", 7) == mpz(22875)

    raises(ValueError, lambda: mpz("123456", base=3))

    assert mpz() == mpz(0)
    assert mpz(Fraction(1,2)) == mpz(0)
    assert mpz(Fraction(-3,2)) == mpz(-1)
    assert mpz(Fraction(3,2)) == mpz(1)
    assert mpz('043') == mpz(43)
    assert mpz('43',0) == mpz(43)
    assert mpz('0o43') == mpz(35)
    assert mpz('0x43') == mpz(67)

    raises(ValueError, lambda: mpz('0x43',10))

    assert mpz('43') == mpz(43)
    assert mpz('1_2') == mpz(12)
    assert mpz('_1_2') == mpz(12)
    assert mpz('1 2') == mpz(12)
    assert mpz(' 1 2') == mpz(12)


@given(integers())
@example(0)
@example(-3)
def test_mpz_conversion_bulk(n):
    assert int(mpz(n)) == n


@settings(max_examples=1000)
@given(integers())
@example(0)
@example(1)
@example(-1)
@example(123456789123456789)
def test_mpz_to_from_binary(n):
    x = mpz(n)
    assert x == from_binary(to_binary(x))


@settings(max_examples=1000)
@given(integers())
@example(0)
@example(1)
@example(-1)
@example(-2)
@example(123)
def test_mpz_hash(n):
    assert hash(mpz(n)) == hash(n)


def test_mpz_ceil():
    a = mpz(123)
    assert math.ceil(a) == a
    assert math.ceil(a) is a


def test_mpz_floor():
    a = mpz(123)
    assert math.floor(a) == a
    assert math.floor(a) is a


def test_mpz_trunc():
    a = mpz(123)
    assert math.trunc(a) == a
    assert math.trunc(a) is a


def test_mpz_round():
    assert round(mpz(123456), 2) == mpz(123456)
    assert round(mpz(123456), -22) == mpz(0)
    assert round(mpz(123456), -2) == mpz(123500)
    assert round(mpz(123456), -1) == mpz(123460)
    assert round(mpz(123455), -1) == mpz(123460)
    assert round(mpz(123454), -1) == mpz(123450)
    assert round(mpz(123445), -1) == mpz(123440)
    assert round(mpz(123445)) == mpz(123445)

    raises(TypeError, lambda: round(mpz(123456),'a'))
    raises(TypeError, lambda: round(mpz(123456),'a',4))


def test_mpz_bool():
    assert bool(mpz(100))
    assert not bool(mpz(0))
    assert bool(mpz(-100))


def test_mpz_random():
    r1 = random_state(42)
    r2 = random_state(42)

    assert mpz_random(r1, 2**88) == mpz(171378365038768291737094841)
    assert mpz_random(r2, 2**88) == mpz(171378365038768291737094841)
    assert mpz_random(r1, 2**88) == mpz(62749575961297098445301393)
    assert mpz_random(r2, 2**88) == mpz(62749575961297098445301393)


def test_mpz_urandomb():
    assert (mpz_urandomb(random_state(42), 64).digits(2) ==
            '1100100011011011101100101001100100111110010111011100101010111001')


def test_mpz_rrandomb():
    assert (mpz_rrandomb(random_state(42), 64).digits(2) ==
            '1111111111111111111111111100000000111111111111111111000000000000')

@mark.skipif(mp_version() < "GMP 6.3.0", reason="requires GMP 6.3.0 or higher")
def test_prev_prime():
    # Imported here as symbol won't exist if mp_version() < 6.3.0
    from gmpy2 import prev_prime
    assert prev_prime(3) == mpz(2)
    assert prev_prime(4) == mpz(3)
    assert prev_prime(5) == mpz(3)
    assert prev_prime(6) == mpz(5)
    assert prev_prime(1000004) == mpz(1000003)
    assert prev_prime(1000033) == mpz(1000003)

    with raises(ValueError):
        prev_prime(-100)

    with raises(ValueError):
        prev_prime(2)

    with raises(TypeError):
        prev_prime('a')

    with raises(TypeError):
        prev_prime(4.5)


def test_mpz_format():
    z1, z2 = mpz(-3), mpz(5)

    assert '{:<5}'.format(z1) == '-3   '
    assert '{:>+5}'.format(z2) == '   +5'

    raises(ValueError, lambda: '{:5+}'.format(z1))

    assert '{:>-4}'.format(z2) == '   5'
    assert '{:<-4}'.format(z1) == '-3  '

    raises(ValueError, lambda: '{:>4-}'.format(z1))
    raises(ValueError, lambda: '{:<4 }'.format(z1))

    assert '{:#x}'.format(z1) == '-0x3'
    assert '{:#o}'.format(z1) == '-0o3'

    raises(ValueError, lambda: '{:>5#}'.format(z1))
    raises(ValueError, lambda: '{:~}'.format(z1))

    a = mpz(123)

    assert str(a) == '123'
    assert repr(a) == 'mpz(123)'
    assert hex(a) == '0x7b'
    assert oct(a) == '0o173'
    assert mpz('1001001011',2) == mpz(587)
    assert bin(mpz('1001001011',2)) == '0b1001001011'
    assert '1001001011' == mpz('1001001011',2).digits(2)
    assert [a.digits(i) for i in range(2,63)] == ['1111011', '11120', '1323',
                                                  '443', '323', '234', '173',
                                                  '146', '123', '102', 'a3',
                                                  '96', '8b', '83', '7b', '74',
                                                  '6f', '69', '63', '5i', '5d',
                                                  '58', '53', '4n', '4j', '4f',
                                                  '4b', '47', '43', '3u', '3r',
                                                  '3o', '3l', '3i', '3f', '3C',
                                                  '39', '36', '33', '30', '2d',
                                                  '2b', '2Z', '2X', '2V', '2T',
                                                  '2R', '2P', '2N', '2L', '2J',
                                                  '2H', '2F', '2D', '2B', '29',
                                                  '27', '25', '23', '21', '1z']

    raises(ValueError, lambda: a.digits(63))

    assert '{}'.format(a) == '123'
    assert '{:d}'.format(a) == '123'
    assert '{:b}'.format(a) == '1111011'
    assert '{:o}'.format(a) == '173'
    assert '{:x}'.format(a) == '7b'
    assert '{:#x}'.format(a) == '0x7b'
    assert '{:#X}'.format(a) == '0X7B'
    assert '{:#o}'.format(a) == '0o173'
    assert '{:#15o}'.format(a) == '          0o173'
    assert '{:<#15o}'.format(a) == '0o173          '
    assert '{:^#15o}'.format(a) == '     0o173     '
    assert '{:>#15o}'.format(a) == '          0o173'
    assert '{:^ #15o}'.format(a) == '     0o173     '
    assert '{:^#15o}'.format(a) == '     0o173     '
    assert '{:^ #16o}'.format(a) == '      0o173     '

    raises(ValueError, lambda: '{:#^16o}'.format(a))

    assert '{:^#16o}'.format(a) == '     0o173      '


def test_mpz_digits():
    z1, z2 = mpz(-3), mpz(15)

    assert z1.digits() == '-3'
    assert z1.digits(2) == '-11'
    assert z1.digits(8) == '-3'
    assert z2.digits(16) == 'f'

    raises(ValueError, lambda: z1.digits(0))
    raises(ValueError, lambda: z1.digits(1))


def test_mpz_abs():
    a = mpz(123)
    b = abs(a)

    assert a is b
    assert abs(-a) == a

    a = mpz(-123)
    b = abs(a)

    assert b == mpz(123)
    assert a is not b
    assert a == mpz(-123)


def test_mpz_add():
    a = mpz(123)
    b = mpz(456)
    c = 12345678901234567890

    assert a+1 == mpz(124)
    assert a+(-1) == mpz(122)
    assert 1+a == mpz(124)
    assert (-1)+a == mpz(122)
    assert a+b == mpz(579)
    assert b+a == mpz(579)

    raises(TypeError, lambda: a+'b')
    raises(TypeError, lambda: 'b'+a)
    assert a+c == 12345678901234568013
    assert c+a == 12345678901234568013

    assert a+True == mpz(124)
    assert a+False == mpz(123)

    assert a + float('Inf') == mpfr('inf')
    assert float('Inf') + a == mpfr('inf')
    assert a + float('-Inf') == mpfr('-inf')
    assert float('-Inf') + a == mpfr('-inf')
    assert is_nan(a + float('nan'))
    assert is_nan(float('nan') + a)


def test_mpz_iadd():
    x = mpz(5)
    x += mpz(6)

    assert x == mpz(11)

    x += 7

    assert x == mpz(18)

    x += -6

    assert x == mpz(12)

    x += 0

    assert x == mpz(12)

    x += mpfr(2.5)

    assert x == mpfr('14.5')


def test_mpz_sub():
    a, b = mpz(123), mpz(456)
    c = 12345678901234567890

    assert a-1 == mpz(122)
    assert a-(-1) == mpz(124)
    assert 1-a == mpz(-122)
    assert (-1)-a == mpz(-124)
    assert a-c == -12345678901234567767
    assert c-a == 12345678901234567767
    assert a-(-c) == 12345678901234568013
    assert (-c)-a == -12345678901234568013
    assert a-b == mpz(-333)
    assert b-a == mpz(333)
    assert a-(-b) == mpz(579)
    assert (-b)-a == mpz(-579)
    assert a-z == mpz(121)
    assert gmpy2.sub(2,1) == mpz(1)

    ctx = gmpy2.context()

    assert ctx.sub(a,b) == a-b
    assert ctx.sub(c,c) == c-c
    assert ctx.sub(1, 1) == mpz(0)
    assert ctx.sub(a, 1) == mpz(122)
    assert ctx.sub(1, a) == mpz(-122)
    assert ctx.sub(a, mpq(0)) == mpq(123,1)
    assert ctx.sub(a, mpfr(0)) == mpfr('123.0')
    assert ctx.sub(a, mpc(0)) == mpc('123.0+0.0j')

    raises(TypeError, lambda: ctx.sub(1))
    raises(TypeError, lambda: ctx.sub(1,2,3))
    raises(TypeError, lambda: a-'b')
    raises(TypeError, lambda: 'b'-a)

    assert a-1 == mpz(122)
    assert a-(-1) == mpz(124)
    assert 1-a == mpz(-122)
    assert (-1)-a == mpz(-124)
    assert a-b == mpz(-333)
    assert b-a == mpz(333)

    raises(TypeError, lambda: a-'b')
    raises(TypeError, lambda: 'b'-a)

    assert a-c == -12345678901234567767
    assert c-a == 12345678901234567767

    assert a - float('Inf') == mpfr('-inf')
    assert float('Inf') - a == mpfr('inf')
    assert a - float('-Inf') == mpfr('inf')
    assert float('-Inf') - a == mpfr('-inf')
    assert is_nan(a - float('nan'))
    assert is_nan(float('nan') - a)


def test_mpz_isub():
    x = mpz(7)
    x -= mpz(1)

    assert x == mpz(6)

    x -= 1

    assert x == mpz(5)

    x -= xmpz(7)

    assert x == mpz(-2)

    x -= -5

    assert x == mpz(3)

    x -= -mpfr(5)

    assert x == mpfr('8.0')


def test_mpz_mul():
    a = mpz(123)
    b = mpz(456)
    c = 12345678901234567890

    assert mpz(2)*z == mpz(4)
    assert gmpy2.mul(2,1) == mpz(2)

    ctx = gmpy2.context()

    assert ctx.mul(a,b) == a*b
    assert ctx.mul(c,c) == c*c
    assert ctx.mul(a, mpq(1)) == mpq(123,1)
    assert ctx.mul(a, mpfr(1)) == mpfr('123.0')
    assert ctx.mul(a, mpc(1)) == mpc('123.0+0.0j')

    raises(TypeError, lambda: ctx.mul(1))
    raises(TypeError, lambda: ctx.mul(1,2,3))

    assert a*b == mpz(56088)
    assert b*a == mpz(56088)
    assert a*0 == mpz(0)
    assert 0*a == mpz(0)
    assert a*123 == mpz(15129)
    assert 123*a == mpz(15129)
    assert a*c == 1518518504851851850470
    assert c*a == 1518518504851851850470

    a = mpz(3)

    assert a*'b' == 'bbb'
    assert 'b'*a == 'bbb'

    assert a*False == mpz(0)

    assert a * float('Inf') == mpfr('inf')
    assert float('Inf') * a == mpfr('inf')
    assert a * float('-Inf') == mpfr('-inf')
    assert float('-Inf') * a == mpfr('-inf')
    assert -a * float('Inf') == mpfr('-inf')
    assert float('Inf') * -a == mpfr('-inf')
    assert -a * float('-Inf') == mpfr('inf')
    assert float('-Inf') * -a == mpfr('inf')
    assert is_nan(a * float('nan'))
    assert is_nan(float('nan') * a)
    assert is_nan(mpz(0) * float('Inf'))
    assert is_nan(mpz(0) * float('-Inf'))
    assert is_nan(float('Inf') * mpz(0))
    assert is_nan(float('-Inf') * mpz(0))


def test_mul_imul():
    x = mpz(2)
    x *= mpz(2)

    assert x == mpz(4)

    x *= 2

    assert x == mpz(8)

    x *= xmpz(3)

    assert x == mpz(24)

    x *= -1

    assert x == mpz(-24)

    x *= mpfr(-0.5)

    assert x == mpfr('12.0')


def test_mpz_divmod():
    a = mpz(123)
    b = mpz(456)
    c = 12345678901234567890
    ctx = gmpy2.get_context()

    raises(TypeError, lambda: divmod(mpz(123),'a'))

    ctx = gmpy2.context()

    raises(TypeError, lambda: ctx.divmod('a',456))
    raises(TypeError, lambda: ctx.divmod(1,2,3))
    raises(ZeroDivisionError, lambda: ctx.divmod(456, 0))
    raises(TypeError, lambda: ctx.divmod(a,mpc(456)))
    raises(TypeError, lambda: divmod(mpz(1), mpc(1,2)))

    assert ctx.divmod(a,b) == (mpz(0), mpz(123))
    assert ctx.divmod(123,456) == (mpz(0), mpz(123))
    assert divmod(mpz(3), z) == (mpz(1), mpz(1))
    assert divmod(z, mpz(3)) == (mpz(0), mpz(2))

    assert divmod(a,b) == (mpz(0), mpz(123))
    assert divmod(b,a) == (mpz(3), mpz(87))

    raises(ZeroDivisionError, lambda: divmod(a,0))
    raises(ZeroDivisionError, lambda: divmod(a, mpz(0)))
    raises(ZeroDivisionError, lambda: divmod(123, mpz(0)))

    assert divmod(b,123) == (mpz(3), mpz(87))
    assert divmod(a,c) == (mpz(0), mpz(123))
    assert divmod(a,int(c)) == (mpz(0), mpz(123))
    assert divmod(a*(c-1),c) == (122, 12345678901234567767)
    assert divmod(a*(c-1),int(c)) == (122, 12345678901234567767)
    assert divmod(a*(c-1),-c) == (mpz(-123), mpz(-123))
    assert divmod(a*(c-1),-int(c)) == (mpz(-123), mpz(-123))
    assert divmod(int(a*(c-1)),-int(c)) == (-123, -123)

    assert divmod(a, mpfr('Inf')) == (mpfr('0.0'), mpfr('123.0'))
    assert divmod(a, mpfr('-Inf')) == (mpfr('-1.0'), mpfr('-inf'))
    assert divmod(-a, mpfr('Inf')) == (mpfr('-1.0'), mpfr('inf'))
    assert divmod(-a, mpfr('-Inf')) == (mpfr('0.0'), mpfr('-123.0'))
    assert all(is_nan(_) for _ in divmod(a, mpfr('nan')))
    assert all(is_nan(_) for _ in divmod(-a, mpfr('nan')))
    assert divmod(mpz(0), mpfr('Inf')) == (mpfr('0.0'), mpfr('0.0'))
    assert divmod(mpz(0), mpfr('-Inf')) == (mpfr('-0.0'), mpfr('-0.0'))
    assert divmod(mpz(0), mpfr('nan'))
    assert all(is_nan(_) for _ in divmod(mpfr('Inf'), a))
    assert all(is_nan(_) for _ in divmod(mpfr('-Inf'), a))
    assert all(is_nan(_) for _ in divmod(mpfr('Inf'), -a))
    assert all(is_nan(_) for _ in divmod(mpfr('-Inf'), -a))
    assert all(is_nan(_) for _ in divmod(mpfr('nan'), a))
    assert all(is_nan(_) for _ in divmod(mpfr('nan'), -a))
    assert all(is_nan(_) for _ in divmod(mpfr('Inf'), mpz(0)))
    assert all(is_nan(_) for _ in divmod(mpfr('-Inf'), mpz(0)))
    assert all(is_nan(_) for _ in divmod(mpfr('nan'), mpz(0)))

    assert divmod(a, mpfr('Inf')) == (mpfr('0.0'), mpfr('123.0'))
    assert divmod(a, mpfr('-Inf')) == (mpfr('-1.0'), mpfr('-inf'))
    assert divmod(-a, mpfr('Inf')) == (mpfr('-1.0'), mpfr('inf'))
    assert divmod(-a, mpfr('-Inf')) == (mpfr('0.0'), mpfr('-123.0'))
    assert all(is_nan(_) for _ in divmod(a, mpfr('nan')))
    assert all(is_nan(_) for _ in divmod(-a, mpfr('nan')))
    assert divmod(mpz(0), mpfr('Inf')) == (mpfr('0.0'), mpfr('0.0'))
    assert divmod(mpz(0), mpfr('-Inf')) == (mpfr('-0.0'), mpfr('-0.0'))
    assert divmod(mpz(0), mpfr('nan'))
    assert all(is_nan(_) for _ in divmod(mpfr('Inf'), a))
    assert all(is_nan(_) for _ in divmod(mpfr('-Inf'), a))
    assert all(is_nan(_) for _ in divmod(mpfr('Inf'), -a))
    assert all(is_nan(_) for _ in divmod(mpfr('-Inf'), -a))
    assert all(is_nan(_) for _ in divmod(mpfr('nan'), a))
    assert all(is_nan(_) for _ in divmod(mpfr('nan'), -a))
    assert all(is_nan(_) for _ in divmod(mpfr('Inf'), mpz(0)))
    assert all(is_nan(_) for _ in divmod(mpfr('-Inf'), mpz(0)))
    assert all(is_nan(_) for _ in divmod(mpfr('nan'), mpz(0)))


def test_mpz_floordiv():
    ctx = gmpy2.get_context()
    a, b = mpz(45), mpz(6)
    r, r2 = mpfr(45), mpfr(3.1)
    q, q2 = mpq(118,18), mpq(3,2)
    c, c2 = mpc(51, 65), mpc(4, 6)

    assert ctx.floor_div(a, 6) == mpz(7)
    assert ctx.floor_div(a, b) == mpz(7)

    raises(ZeroDivisionError, lambda: ctx.floor_div(a, 0))
    raises(ZeroDivisionError, lambda: ctx.floor_div(a, mpz(0)))

    assert ctx.floor_div(45, b) == mpz(7)

    raises(ZeroDivisionError, lambda: ctx.floor_div(45, 0))
    raises(ZeroDivisionError, lambda: ctx.floor_div(45, mpz(0)))
    raises(TypeError, lambda: ctx.floor_div())
    raises(TypeError, lambda: gmpy2.floor_div(4,5,6))

    assert a // b == mpz(7)

    raises(ZeroDivisionError, lambda: a // 0)
    raises(ZeroDivisionError, lambda: a // mpz(0))

    assert ctx.floor_div(a, q2) == mpz(30)
    assert ctx.floor_div(a, r2) == mpfr('14.0')

    raises(TypeError, lambda: ctx.floor_div(a, c))

    assert a // b == mpz(7)
    assert a // q == mpz(6)
    assert a // r2 == mpfr('14.0')

    raises(TypeError, lambda: a // c2)
    raises(TypeError, lambda: a // 'not')

    a = mpz(123)
    b = mpz(456)
    c = 12345678901234567890

    assert a//b == mpz(0)
    assert b//a == mpz(3)
    assert (a*b)//b == mpz(123)
    assert (a*b)//a == mpz(456)

    raises(ZeroDivisionError, lambda: a//0)

    assert c//a == 100371373180768844
    assert a**10//c == mpz(64)
    assert a // z == mpz(61)

    assert a//True == mpz(123)


def test_mpz_ifloordiv():
    x = mpz(49)
    x //= mpz(3)

    assert x == mpz(16)

    x //= xmpz(3)

    assert x == mpz(5)

    x //= 2

    assert x == mpz(2)

    with raises(ZeroDivisionError):
        x //= mpz(0)
    with raises(ZeroDivisionError):
        x //= 0

    assert x == mpz(2)

    x //= mpfr(-0.5)

    assert x == mpfr('-4.0')

    x = mpz(11)
    x //= -5

    assert x == mpz(-3)


def test_mpz_mod():
    a = mpz(123)
    b = mpz(456)
    c = 12345678901234567890
    ctx = gmpy2.get_context()

    assert a % b == mpz(123)
    assert b % a == mpz(87)
    assert gmpy2.mod(b, a) == mpz(87)
    assert ctx.mod(b, a) == mpz(87)
    assert a % z == mpz(1)

    raises(ZeroDivisionError, lambda: a % mpz(0))
    raises(ZeroDivisionError, lambda: a % 0)
    raises(ZeroDivisionError, lambda: 14 % mpz(0))
    raises(ZeroDivisionError, lambda: gmpy2.mod(14, mpz(0)))
    raises(ZeroDivisionError, lambda: gmpy2.mod(124, 0))
    raises(ZeroDivisionError, lambda: gmpy2.mod(b, mpz(0)))
    raises(ZeroDivisionError, lambda: ctx.mod(b, mpz(0)))
    raises(TypeError, lambda: gmpy2.mod(124, 'str'))
    raises(TypeError, lambda: a % 'str')

    assert gmpy2.mod(124, mpz(5)) == mpz(4)
    assert z % mpq(1,2) == mpq(0,1)
    assert a % mpq(2,3) == mpq(1,3)


def test_mpz_imod():
    x = mpz(45)
    x %= mpz(18)

    assert x == mpz(9)

    x %= xmpz(2)

    assert x == mpz(1)

    x = mpz(40)
    x %= 21

    assert x == mpz(19)

    with raises(ZeroDivisionError):
        x %= 0

    assert x == mpz(19)

    with raises(ZeroDivisionError):
        x %= mpz(0)

    x %= -9

    assert x == mpz(-8)

    x %= mpfr(10)

    assert x == mpfr('2.0')


def test_mpz_truediv():
    a = mpz(123)
    b = mpz(456)
    c = 12345678901234567890
    ctx = gmpy2.get_context()

    assert a/b == mpfr('0.26973684210526316')
    assert gmpy2.div(a, b) == mpfr('0.26973684210526316')
    assert ctx.div(a, b) == mpfr('0.26973684210526316')
    assert b/a == mpfr('3.7073170731707319')

    raises(ZeroDivisionError, lambda: a/0)

    assert a/0.0 == mpfr('inf')
    assert a / z == mpfr('61.5')

    raises(TypeError, lambda: ctx.div(a, b, 5))
    raises(TypeError, lambda: ctx.div(a, 'str'))
    raises(TypeError, lambda: a / 'str')

    with gmpy2.context(rational_division=True):
        assert mpz(1)/mpz(2) == mpq(1, 2)

    assert a / float('Inf') == mpfr('0.0')
    assert -a / float('Inf') == mpfr('-0.0')
    assert float('Inf') / a == mpfr('inf')
    assert float('Inf') / -a == mpfr('-inf')
    assert a / float('-Inf') == mpfr('-0.0')
    assert -a / float('-Inf') == mpfr('0.0')
    assert float('-Inf') / a == mpfr('-inf')
    assert float('-Inf') / -a == mpfr('inf')
    assert is_nan(a / float('nan'))
    assert is_nan(float('nan') / a)
    assert is_nan(float('nan') / mpz(0))
    assert is_nan(float('nan') / mpz(0))

    assert a / mpfr('Inf') == mpfr('0.0')
    assert -a / mpfr('Inf') == mpfr('-0.0')
    assert mpfr('Inf') / a == mpfr('inf')
    assert mpfr('Inf') / -a == mpfr('-inf')
    assert a / mpfr('-Inf') == mpfr('-0.0')
    assert -a / mpfr('-Inf') == mpfr('0.0')
    assert mpfr('-Inf') / a == mpfr('-inf')
    assert mpfr('-Inf') / -a == mpfr('inf')
    assert is_nan(a / mpfr('nan'))
    assert is_nan(mpfr('nan') / a)
    assert is_nan(mpfr('nan') / mpz(0))
    assert is_nan(mpfr('nan') / mpz(0))


def test_mpz_pow():
    z1, z2 = mpz(5), mpz(2)
    ctx = gmpy2.get_context()

    assert z1 ** z2 == mpz(25)
    assert ctx.pow(z1, z2) == mpz(25)
    assert z1 ** -z2 == mpfr('0.040000000000000001')
    assert z1 ** 0 == mpz(1)
    assert mpz(0) ** 32 == mpz(0)
    assert mpz(-1) ** 32 == mpz(1)
    assert mpz(1) ** mpz(324) == mpz(1)
    assert mpz(0) ** 0 == mpz(1)
    assert mpz(-1) ** 3 == mpz(-1)
    assert z1 ** 2 == pow(z1, 2)
    assert pow(z1, 2, 19) == mpz(6)
    assert pow(z1, -2, 19) == mpz(16)

    raises(ValueError, lambda: pow(mpz(0), -2, 19))

    assert pow(z1, 2, -19) == mpz(-13)

    raises(ValueError, lambda: pow(5, 2, 0))
    raises(TypeError, lambda: ctx.pow(z1, 'invalid'))

    a = mpz(123)
    b = mpz(456)

    assert pow(a,10) == 792594609605189126649
    assert pow(a,7,b) == mpz(99)


def test_mpz_ipow():
    x = mpz(5)
    x **= mpz(2)

    assert x == mpz(25)

    x **= xmpz(2)

    assert x == mpz(625)

    x **= -2

    assert x == mpfr('2.5600000000000001e-06')

    x = mpz(625)
    x **= 2

    assert x == mpz(390625)

    x **= mpfr(2)

    assert x == mpfr('152587890625.0')

    x = mpz(390625)
    x **= mpfr(-2)

    assert x == mpfr('6.5535999999999999e-12')


def test_lucasu():
    assert gmpy2.lucasu(2,4,0) == mpz(0)
    assert gmpy2.lucasu(2,4,1) == mpz(1)

    raises(ValueError, lambda: gmpy2.lucasu(2,1,1))

    assert gmpy2.lucasu(2,4,8) == mpz(128)

    raises(TypeError, lambda: gmpy2.lucasu('a',4,8))
    raises(TypeError, lambda: gmpy2.lucasu(2,'b',8))
    raises(TypeError, lambda: gmpy2.lucasu(2,4,None))
    raises(ValueError, lambda: gmpy2.lucasu(mpz(2), mpz(1), mpz(7)))

    assert gmpy2.lucasu_mod(3,2,0,7) == mpz(0)
    assert gmpy2.lucasu_mod(3,2,5,7) == mpz(3)
    assert gmpy2.lucasu_mod(3,2,555,777777777) == mpz(387104641)

    raises(ValueError, lambda: gmpy2.lucasu_mod(2,1,555,777777777))

    raises(ValueError, lambda: gmpy2.lucasv(2,1,4))
    raises(TypeError, lambda: gmpy2.lucasv('a',1,2))
    raises(TypeError, lambda: gmpy2.lucasv(4,'b',2))
    raises(TypeError, lambda: gmpy2.lucasv(4,3,'c'))

    assert gmpy2.lucasv(4,3,0) == mpz(2)
    assert gmpy2.lucasv(4,3,7) == mpz(2188)
    assert gmpy2.lucasv(4,3,8) == mpz(6562)

    assert gmpy2.lucasv_mod(4,3,0,2) == mpz(0)
    assert gmpy2.lucasv_mod(4,3,0,3) == mpz(2)
    assert gmpy2.lucasv_mod(4,3,55,123456) == mpz(35788)
    assert gmpy2.lucasv_mod(4,3,56,123456) == mpz(107362)
    assert gmpy2.lucasv_mod(4,3,57,123456) == mpz(75172)


def test_mpz_attributes():
    a = mpz(123)

    assert a.numerator == mpz(123)
    assert a.denominator == mpz(1)

    assert a.real == mpz(123)
    assert a.imag == mpz(0)


def test_mpz_conjugate():
    a = mpz(123)

    assert a.conjugate() == a


def test_mpz_invert():
    a = mpz(123)

    assert ~a == mpz(-124)


def test_mpz_and():
    a = mpz(123)
    b = mpz(456)

    assert a&b == mpz(72)
    assert a&int(b) == mpz(72)
    assert int(a)&b == mpz(72)

    raises(TypeError, lambda: a&mpq(1))


def test_mpz_and_operator():
    assert (mpz(0) and mpz(7)) == mpz(0)
    assert (mpz(7) and mpz(0)) == mpz(0)
    assert (mpz(7) and mpz(5)) == mpz(5)
    assert (mpz(0) and 5) == mpz(0)
    assert (mpz(7) and 5.2) == 5.2
    assert (mpz(7) and None) is None


def test_mpz_iand():
    x = mpz(7)
    x &= mpz(5)

    assert x == mpz(5)

    x &= xmpz(4)

    assert x == mpz(4)

    x &= 9

    assert x == mpz(0)

    x = mpz(4)
    x &= 12

    assert x == mpz(4)

    with raises(TypeError):
        x &= mpfr(4)


def test_mpz_or():
    a = mpz(123)
    b = mpz(456)

    assert a|b == mpz(507)
    assert a|int(b) == mpz(507)
    assert int(a)|b == mpz(507)

    raises(TypeError, lambda: a|mpq(1))


def test_mpz_ior():
    x = mpz(0)
    x |= mpz(1)

    assert x == mpz(1)

    x |= mpz(0)

    assert x == mpz(1)

    x = mpz(0)

    x |= mpz(0)

    assert x == mpz(0)

    x |= 5

    assert x == mpz(5)

    with raises(TypeError):
        x |= mpfr(3)


def test_mpz_xor():
    a = mpz(123)
    b = mpz(456)

    assert a^b == mpz(435)
    assert a^int(b) == mpz(435)
    assert int(a)^b == mpz(435)

    raises(TypeError, lambda: a^mpq(1))


def test_mpz_ixor():
    x = mpz(1)
    x ^= mpz(0)

    assert x == mpz(1)

    x ^= xmpz(1)

    assert x == mpz(0)

    x ^= 1

    assert x == mpz(1)

    with raises(TypeError):
        x ^= mpfr(0)


def test_mpz_lshift():
    a = mpz(123)

    assert a<<1 == mpz(246)
    assert int(a)<<mpz(1) == mpz(246)

    raises(OverflowError, lambda: a<<-1)

    assert a<<0 == mpz(123)

    raises(TypeError, lambda: "a" << a)


def test_mpz_rshift():
    a = mpz(123)

    assert a>>1 == mpz(61)
    assert int(a)>>mpz(1) == mpz(61)

    raises(OverflowError, lambda: a>>-2)

    assert a>>0 == mpz(123)

    raises(TypeError, lambda: "a" >> a)


def test_mpz_ilshift_irshift():
    x = mpz(63)
    x >>= mpz(63)

    assert x == mpz(0)

    x = mpz(63)
    x >>= mpz(1)

    assert x == mpz(31)

    x >>= xmpz(2)

    assert x == mpz(7)

    x >>= 1

    assert x == mpz(3)

    x <<= mpz(2)

    assert x == mpz(12)

    x <<= mpz(1)

    assert x == mpz(24)

    x <<= 0

    assert x == mpz(24)

    with raises(TypeError):
        x >>= mpfr(2)
    with raises(TypeError):
        x <<= mpfr(2)
    with raises(OverflowError):
        x >>= -1
    with raises(OverflowError):
        x <<= -5


def test_mpz_index():
    a = mpz(123)
    b = mpz(456)

    assert range(333)[a] == 123

    raises(IndexError, lambda: range(333)[b])


def test_mpz_seq():
    a = mpz(10)

    assert a[1] == 1
    assert a[-1] == 1
    assert a[-2] == 0
    assert a[0:2] == mpz(2)
    assert a[0:3] == mpz(2)
    assert a[0:3:-1] == mpz(0)

    raises(IndexError, lambda: a[111111111111111111111])
    raises(TypeError, lambda: a["spam"])


def test_mpz_len():
    assert len(mpz(0)) == 1
    assert len(mpz(1)) == 1
    assert len(mpz(17)) == 5


def test_mpz_num_digits():
    assert mpz(123456).num_digits() == 6
    assert mpz(123456).num_digits(2) == 17

    raises(ValueError, lambda: mpz(123456).num_digits(-1))
    raises(OverflowError, lambda: mpz(123456).num_digits(9999999999999999999999999999999999))
    raises(ValueError, lambda: mpz(123456).num_digits(100))
    raises(TypeError, lambda: gmpy2.num_digits(123456,-1,7))
    raises(ValueError, lambda: gmpy2.num_digits(123456,-1))
    raises(TypeError, lambda: gmpy2.num_digits('123456'))
    raises(ValueError, lambda: gmpy2.num_digits(123456,100))
    raises(TypeError, lambda: gmpy2.num_digits(123456,'a'))

    assert gmpy2.num_digits(123456) == 6
    assert gmpy2.num_digits(123456,2) == 17


def test_mpz_is_square():
    raises(TypeError, lambda: gmpy2.is_square('a'))

    assert gmpy2.is_square(mpz(9))
    assert not gmpy2.is_square(10)
    assert mpz(16).is_square()
    assert not mpz(17).is_square()


def test_mpz_is_divisible():
    raises(TypeError, lambda: gmpy2.is_divisible())
    raises(TypeError, lambda: gmpy2.is_divisible('a',2))
    raises(TypeError, lambda: gmpy2.is_divisible(2,'a'))

    assert gmpy2.is_divisible(12,2)
    assert not gmpy2.is_divisible(12,7)
    assert mpz(12).is_divisible(2)
    assert not mpz(12).is_divisible(7)
    assert gmpy2.is_divisible(mpz(123456789123456789123456789),
                              123456789123456789123456789)
    assert not gmpy2.is_divisible(mpz(1234567891234567891234567897),
                                  123456789123456789123456789)

    raises(TypeError, lambda: mpz(12).is_divisible('a'))

    assert mpz(123456789123456789123456789).is_divisible(123456789123456789123456789)
    assert not mpz(1234567891234567891234567897).is_divisible(123456789123456789123456789)


def test_mpz_is_congruent():
    raises(TypeError, lambda: gmpy2.is_congruent(1))
    raises(TypeError, lambda: gmpy2.is_congruent(1,'a',3))

    assert gmpy2.is_congruent(7*3+2, 7*11+2, 7)
    assert not gmpy2.is_congruent(7*3+2, 7*11+5, 7)

    raises(TypeError, lambda: mpz(7*3+2).is_congruent(1))
    raises(TypeError, lambda: mpz(7*3+2).is_congruent('a',7))

    assert mpz(7*3+2).is_congruent(7*11+2,7)
    assert not mpz(7*3+2).is_congruent(7*11+5,7)


def test_mpz_is_power():
    a = mpz(123)

    raises(TypeError, lambda: gmpy2.is_power())
    raises(TypeError, lambda: gmpy2.is_power('a'))

    assert not a.is_power()
    assert mpz(123**11).is_power()
    assert not gmpy2.is_power(a)
    assert gmpy2.is_power(99*99*99)
    assert not gmpy2.is_power(99*98)


def test_mpz_is_prime():
    raises(OverflowError, lambda: gmpy2.is_prime(3,-3))
    raises(TypeError, lambda: gmpy2.is_prime())
    raises(TypeError, lambda: gmpy2.is_prime(1,2,3))
    raises(TypeError, lambda: gmpy2.is_prime('a'))

    assert not gmpy2.is_prime(12345)
    assert gmpy2.is_prime(80**81 + 81**80)
    assert gmpy2.is_prime(80**81 + 81**80, 10000)

    raises(TypeError, lambda: mpz(129).is_prime(1,2))
    raises(OverflowError, lambda: mpz(129).is_prime(-7))

    assert not mpz(129).is_prime(10000)
    assert mpz(80**81 + 81**80).is_prime()
    assert not mpz(1234567890).is_prime()
    assert not mpz(-3).is_prime()
    assert not gmpy2.is_prime(-3)


def test_mpz_is_probab_prime():
    raises(OverflowError, lambda: gmpy2.is_probab_prime(3,-3))
    raises(TypeError, lambda: gmpy2.is_probab_prime())
    raises(TypeError, lambda: gmpy2.is_probab_prime(1,2,3))
    raises(TypeError, lambda: gmpy2.is_probab_prime('a'))

    assert gmpy2.is_probab_prime(71) == 2
    assert gmpy2.is_probab_prime(12345) == 0
    assert gmpy2.is_probab_prime(80**81 + 81**80) == 1
    assert gmpy2.is_probab_prime(80**81 + 81**80, 10000) == 1

    raises(TypeError, lambda: mpz(129).is_probab_prime(1,2))
    raises(OverflowError, lambda: mpz(129).is_probab_prime(-7))

    assert mpz(71).is_probab_prime(71) == 2
    assert mpz(129).is_probab_prime(10000) == 0
    assert mpz(80**81 + 81**80).is_probab_prime() == 1
    assert mpz(1234567890).is_probab_prime() == 0
    assert mpz(-3).is_probab_prime() == 0
    assert gmpy2.is_probab_prime(-3) == 0


def test_mpz_is_even():
    a = mpz(123)
    b = mpz(456)

    raises(TypeError, lambda: gmpy2.is_even('a'))

    assert not gmpy2.is_even(a)
    assert gmpy2.is_even(b)
    assert not a.is_even()
    assert b.is_even()
    assert not gmpy2.is_even(11)
    assert gmpy2.is_even(14)


def test_mpz_is_odd():
    a = mpz(123)
    b = mpz(456)

    raises(TypeError, lambda: gmpy2.is_odd('a'))

    assert gmpy2.is_odd(a)
    assert not gmpy2.is_odd(b)
    assert a.is_odd()
    assert not b.is_odd()
    assert gmpy2.is_odd(11)
    assert not gmpy2.is_odd(14)


def test_mpz_bit_length():
    assert mpz(0).bit_length() == 0
    assert mpz(1).bit_length() == 1
    assert mpz(5).bit_length() == 3
    assert mpz(8).bit_length() == 4
    assert gmpy2.bit_length(mpz(10**30)) == 100
    assert gmpy2.bit_length(56) == 6

    raises(TypeError, lambda: gmpy2.bit_length(mpfr(4.0)))


def test_mpz_bit_mask():
    assert gmpy2.bit_mask(mpz(0)) == mpz(0)
    assert gmpy2.bit_mask(mpz(4)) == mpz(15)
    assert gmpy2.bit_mask(mpz(3)) == mpz(7)
    assert gmpy2.bit_mask(mpz(16)) == mpz(65535)
    assert gmpy2.bit_mask(8) == mpz(255)

    raises(OverflowError, lambda: gmpy2.bit_mask(-1))


def test_mpz_bit_scan0():
    assert mpz(6).bit_scan0() == 0
    assert mpz(7).bit_scan0() == 3
    assert mpz(8).bit_scan0(2) == 2
    assert mpz(7).bit_scan0(2) == 3

    raises(OverflowError, lambda: mpz(7).bit_scan0(-2))

    assert gmpy2.bit_scan0(mpz(7), 2) == 3
    assert gmpy2.bit_scan0(mpz(8), 2) == 2
    assert gmpy2.bit_scan0(8) == 0

    raises(TypeError, lambda: gmpy2.bit_scan0())
    raises(TypeError, lambda: gmpy2.bit_scan0(mpz(7), 2.5))
    raises(TypeError, lambda: gmpy2.bit_scan0(mpz(7), 2, 5))
    raises(TypeError, lambda: gmpy2.bit_scan0(7.5, 0))
    raises(OverflowError, lambda: gmpy2.bit_scan0(8, -2))

    assert gmpy2.bit_scan0(mpz(-1), 1) is None
    assert mpz(-1).bit_scan0(1) is None


def test_mpz_bit_scan1():
    assert mpz(7).bit_scan1() == 0
    assert mpz(8).bit_scan1() == 3
    assert mpz(7).bit_scan1(2) == 2

    raises(OverflowError, lambda: mpz(7).bit_scan1(-2))

    assert gmpy2.bit_scan1(7) == 0
    assert gmpy2.bit_scan1(8) == 3
    assert gmpy2.bit_scan1(7, 2) == 2

    raises(TypeError, lambda: gmpy2.bit_scan1(mpz(7), 2, 5))
    raises(TypeError, lambda: gmpy2.bit_scan1())
    raises(TypeError, lambda: gmpy2.bit_scan1(mpz(6), 2.5))
    raises(TypeError, lambda: gmpy2.bit_scan1(7.5, 0))
    raises(OverflowError, lambda: gmpy2.bit_scan1(8, -1))

    assert gmpy2.bit_scan1(mpz(1), 1) is None
    assert mpz(1).bit_scan1(1) is None


def test_mpz_bit_test():
    assert mpz(7).bit_test(2)
    assert not mpz(8).bit_test(2)
    assert not mpz(-8).bit_test(2)

    raises(OverflowError, lambda: mpz(8).bit_test(-2))

    assert gmpy2.bit_test(mpz(7), 2)
    assert not gmpy2.bit_test(mpz(8), 2)

    raises(TypeError, lambda: gmpy2.bit_test())
    raises(TypeError, lambda: gmpy2.bit_test(mpz(7), 2.5))
    raises(TypeError, lambda: gmpy2.bit_test(7.5, 2))
    raises(OverflowError, lambda: gmpy2.bit_test(8, -2))


def test_mpz_bit_clear():
    assert mpz(7).bit_clear(0) == mpz(6)
    assert mpz(7).bit_clear(2) == mpz(3)
    assert mpz(8).bit_clear(2) == mpz(8)

    raises(OverflowError, lambda: mpz(8).bit_clear(-1))

    assert gmpy2.bit_clear(4, 2) == mpz(0)

    raises(TypeError, lambda: gmpy2.bit_clear())
    raises(TypeError, lambda: gmpy2.bit_clear(7.2, 2))
    raises(TypeError, lambda: gmpy2.bit_clear(mpz(4), 2.5))
    raises(OverflowError, lambda: gmpy2.bit_clear(4, -2))


def test_mpz_bit_set():
    assert mpz(4).bit_set(0) == mpz(5)
    assert mpz(7).bit_set(3) == mpz(15)
    assert mpz(0).bit_set(2) == mpz(4)

    raises(OverflowError, lambda: mpz(0).bit_set(-2))

    assert gmpy2.bit_set(8, 1) == mpz(10)

    raises(TypeError, lambda: gmpy2.bit_set(0))
    raises(TypeError, lambda: gmpy2.bit_set())
    raises(TypeError, lambda: gmpy2.bit_set(8.5, 1))
    raises(TypeError, lambda: gmpy2.bit_set(8, 1.5))
    raises(OverflowError, lambda: gmpy2.bit_set(8, -1))


def test_mpz_bit_flip():
    assert mpz(4).bit_flip(2) == mpz(0)
    assert mpz(4).bit_flip(1) == mpz(6)
    assert mpz(0).bit_flip(3) == mpz(8)

    raises(OverflowError, lambda: mpz(5).bit_flip(-3))

    assert gmpy2.bit_flip(mpz(7), mpz(1)) == mpz(5)
    assert gmpy2.bit_flip(mpz(7), 2) == mpz(3)

    raises(TypeError, lambda: gmpy2.bit_flip())
    raises(TypeError, lambda: gmpy2.bit_flip(4.5, 2))
    raises(TypeError, lambda: gmpy2.bit_flip(4, 2.5))
    raises(OverflowError, lambda: gmpy2.bit_flip(mpz(7), -2))


def test_mpz_bit_count():
    a = mpz(10009)

    assert a.bit_count() == 7
    assert gmpy2.popcount(a) == 7
    assert gmpy2.bit_count(a) == 7

    a = -a

    assert a == mpz(-10009)
    assert a.bit_count() == 7
    assert gmpy2.bit_count(a) == 7
    assert gmpy2.popcount(a) == -1

    raises(TypeError, lambda: gmpy2.bit_count('spam'))


def test_mpz_popcount():
    assert gmpy2.popcount(-65) == -1
    assert gmpy2.popcount(7) == 3
    assert gmpy2.popcount(8) == 1
    assert gmpy2.popcount(15) == 4
    assert gmpy2.popcount(mpz(0)) == 0

    raises(TypeError, lambda: gmpy2.popcount(4.5))


def test_mpz_hamdist():
    assert gmpy2.hamdist(mpz(5), mpz(7)) == 1
    assert gmpy2.hamdist(mpz(0), mpz(7)) == 3
    assert gmpy2.hamdist(mpz(0), 7) == 3

    raises(TypeError, lambda: gmpy2.hamdist(mpq(14,2), 5))
    raises(TypeError, lambda: gmpy2.hamdist(5,6,5))


def test_issue_339():
    samples = map(mpz, [13157547707030902665, 1070317427780135395,
                        18019609787501108695, 3978762157568107671,
                        14444587867185512177])
    assert all((2*q).is_divisible(q) for q in samples)


def test_issue_312():
    assert not is_prime(-7)
    assert not is_prime(mpz(-7))
    assert not is_prime(1 - 2**4423)
    assert all(not is_prime(-a) for a in range(8))
    assert next_prime(-8) == 2



import math
import numbers
import pickle
import sys
from decimal import Decimal
from fractions import Fraction

import pytest
from hypothesis import example, given, settings
from hypothesis.strategies import fractions, integers
from supportclasses import a, b, c, d, q, z

import gmpy2
from gmpy2 import (cmp, cmp_abs, from_binary, is_nan, mpc, mpfr, mpq, mpz,
                   to_binary, xmpz)


def test_mpz_constructor():
    assert mpq('1/1') == mpq(1,1)
    assert mpq('1_/_1') == mpq(1,1)


def test_mpq_as_integer_ratio():
    assert mpq(2, 3).as_integer_ratio() == (mpz(2), mpz(3))


def test_mpq_numbers_abc():
    assert isinstance(mpq(1, 3), numbers.Rational)


def test_mpq_pickling():
    for proto in range(pickle.HIGHEST_PROTOCOL + 1):
        for x in [mpq(1234, 6789), mpq(-1234, 6789), mpq(0, 1)]:
            assert pickle.loads(pickle.dumps(x, protocol=proto)) == x


def test_mpq_from_float():
    assert mpq.from_float(3.2) == mpq(3602879701896397, 1125899906842624)


def test_mpq_from_Decimal():
    assert mpq(Decimal("5e-3")) == mpq(5, 1000)
    assert mpq(Decimal(1)) == mpq(1)  # issue 327
    assert mpq(Decimal('0.6')) == mpq(3, 5)
    assert mpq.from_decimal(Decimal("5e-3")) == mpq(5, 1000)


def test_mpq_cmp():
    assert cmp(mpq(1,2), mpq(1,2)) == 0
    assert cmp(mpq(-1,2), mpq(1,2)) == -1
    assert cmp(mpq(1,2), mpq(-1,2)) == 1
    assert cmp(0, mpq(1,2)) == -1
    assert cmp(mpq(1,2), 0) == 1

    assert cmp_abs(mpq(1,2), mpq(1,3)) == 1
    assert cmp_abs(mpq(-1,2), mpq(1,3)) == 1
    assert cmp_abs(mpq(1,2), mpq(-1,3)) == 1
    assert cmp_abs(mpq(1,4), mpq(1,3)) == -1
    assert cmp_abs(mpq(-1,4), mpq(1,3)) == -1
    assert cmp_abs(mpq(1,4), mpq(-1,3)) == -1

    assert cmp(mpq(3,2), q) == 0
    assert cmp(q, mpq(3,5)) == 1


def test_mpq_comparisons():
    from supportclasses import q

    assert mpq(3,2) == q
    assert (q == mpq(3,5)) is False

    a = mpz(123)
    q = mpq(4, 5)

    assert (q == a, q != a, q > a, q >= a, q < a, q <= a) == (False, True, False, False, True, True)
    assert (mpq(246,2) != a) is False

    f = float(0.7)

    assert (q == f, q != f, q > f, q >= f, q < f, q <= f) == (False, True, True, True, False, False)

    f = float('nan')

    assert (q == f, q != f, q > f, q >= f, q < f, q <= f) == (False, True, False, False, False, False)

    f = float('inf')

    assert (q == f, q != f, q > f, q >= f, q < f, q <= f) == (False, True, False, False, True, True)

    f = -f

    assert (q == f, q != f, q > f, q >= f, q < f, q <= f) == (False, True, True, True, False, False)


def test_mpq_conversion():
    x = mpq(a)
    assert isinstance(x, mpq)
    assert 2*x == 3
    assert mpq(z, z) == mpq(1, 1)

    class Three:
        def __mpz__(self): return mpz(3)

    assert mpq(z, Three()) == mpq(2,3)

    pytest.raises(TypeError, lambda: mpq(b))
    pytest.raises(TypeError, lambda: mpq(c))
    pytest.raises(TypeError, lambda: mpq(d))

    assert mpq('2/3') == mpq(2,3)
    assert mpq(b'2/3') == mpq(2,3)

    pytest.raises(ValueError, lambda: mpq('2,3'))
    pytest.raises(ValueError, lambda: mpq('2/a'))

    assert mpq('2.3') == mpq(23,10)

    assert pytest.raises(ValueError, lambda: mpq('2.3/10'))

    assert mpq(4.5) == mpq(9,2)

    pytest.raises(OverflowError, lambda: mpq(float('inf')))

    assert mpq(xmpz(15)) == mpq(15,1)
    assert mpq(mpfr(4.5)) == mpq(9,2)

    pytest.raises(TypeError, lambda: mpq(dict()))

    assert float(mpq(1,2)) == 0.5
    assert int(mpq(15,2)) == 7


def test_mpq_round():
    pytest.raises(TypeError, lambda: round(mpq(7,3), 4.5))

    q = mpq('4/5')

    assert round(mpq('7/2')) == mpz(4)
    assert round(q, 4) == mpq(4,5)

    pytest.raises(TypeError, lambda: round(q, 4, 2))


@settings(max_examples=1000)
@given(integers(), integers(min_value=1))
@example(0, 1)
@example(1, 1)
@example(-1, 2)
@example(123456789123456789, 9876)
def test_mpq_to_from_binary(p, q):
    x = mpq(p,q)
    assert x == from_binary(to_binary(x))


@settings(max_examples=1000)
@given(fractions())
@example(Fraction(0))
@example(Fraction(-1))
@example(Fraction(-2))
@example(Fraction(15432, 125))
@example(Fraction(1, sys.hash_info.modulus))
@example(Fraction(-1, sys.hash_info.modulus))
def test_mpq_hash(q):
    assert hash(mpq(q)) == hash(q)


def test_mpq_digits():
    q = mpq(2/3)

    assert q.digits() == '6004799503160661/9007199254740992'
    assert q.digits(16) == '0x15555555555555/0x20000000000000'

    pytest.raises(TypeError, lambda: q.digits(16, 5))
    pytest.raises(ValueError, lambda: q.digits(0))


def test_mpq_abs():
    a = mpq(12,7)
    b = abs(a)

    assert a is b

    a = mpq(-12,7)
    b = abs(a)

    assert b == mpq(12,7)
    assert a == mpq(-12,7)


def test_mpq_add():
    a = mpq(3,11)

    assert a + float('Inf') == mpfr('inf')
    assert float('Inf') + a == mpfr('inf')
    assert a + float('-Inf') == mpfr('-inf')
    assert float('-Inf') + a == mpfr('-inf')
    assert is_nan(a + float('nan'))
    assert is_nan(float('nan') + a)

    assert a + mpfr('Inf') == mpfr('inf')
    assert mpfr('Inf') + a == mpfr('inf')
    assert a + mpfr('-Inf') == mpfr('-inf')
    assert mpfr('-Inf') + a == mpfr('-inf')
    assert is_nan(a + mpfr('nan'))
    assert is_nan(mpfr('nan') + a)


def test_mpq_sub():
    assert mpq(1,2) - Fraction(3,2) == mpq(-1,1)
    assert Fraction(1,2) - mpq(3,2) == mpq(-1,1)
    assert mpq(1,2) - mpq(3,2) == mpq(-1,1)
    assert mpq(1,2) - 0 == mpq(1,2)
    assert mpq(1,2) - mpz(1) == mpq(-1,2)
    assert mpq(1,2) + (-1) == mpq(-1,2)
    assert 1 - mpq(1,2) == mpq(1,2)
    assert mpz(1) - mpq(1,2) == mpq(1,2)
    assert mpq(1,2) - mpz(1) == mpq(-1,2)
    assert mpq(1,1) - mpc(0) == mpc('1.0+0.0j')
    assert mpc(0) - mpq(1,1) == mpc('-1.0+0.0j')
    assert mpq(1,2) - z == mpq(-3,2)
    assert mpq(1,2) - q ==mpq(-1,1)

    ctx = gmpy2.context()

    assert ctx.sub(mpq(1,2), mpq(3,2)) == mpq(-1,1)
    assert ctx.sub(mpq(1,2), Fraction(3,2)) == mpq(-1,1)
    assert ctx.sub(Fraction(1,2), mpq(3,2)) == mpq(-1,1)
    assert ctx.sub(Fraction(1,2), Fraction(3,2)) == mpq(-1,1)

    pytest.raises(TypeError, lambda: ctx.sub(1,'a'))
    pytest.raises(TypeError, lambda: mpq(1,2) - 'a')
    pytest.raises(TypeError, lambda: 'a' - mpq(1,2))

    a = mpq(3,11)
    b = mpq(1,2)
    c = Fraction(5,7)

    assert a-b == mpq(-5,22)
    assert b-a == mpq(5,22)
    assert a-1 == mpq(-8,11)
    assert 1-a == mpq(8,11)
    assert a-c == mpq(-34,77)
    assert c-a == mpq(34,77)
    assert a-a == mpq(0,1)
    assert a-mpz(123456) == mpq(-1358013,11)
    assert mpz(-123456)-a == mpq(-1358019,11)

    pytest.raises(TypeError, lambda: a-'b')
    pytest.raises(TypeError, lambda: 'b'-a)

    assert a - float('Inf') == mpfr('-inf')
    assert float('Inf') - a == mpfr('inf')
    assert a - float('-Inf') == mpfr('inf')
    assert float('-Inf') - a == mpfr('-inf')
    assert is_nan(a - float('nan'))
    assert is_nan(float('nan') - a)

    assert a - mpfr('Inf') == mpfr('-inf')
    assert mpfr('Inf') - a == mpfr('inf')
    assert a - mpfr('-Inf') == mpfr('inf')
    assert mpfr('-Inf') - a == mpfr('-inf')
    assert is_nan(a - mpfr('nan'))
    assert is_nan(mpfr('nan') - a)


def test_mpq_mul():
    assert mpq(1,2) * Fraction(3,2) == mpq(3,4)
    assert Fraction(1,2) * mpq(3,2) == mpq(3,4)
    assert mpq(1,2) * mpq(3,2) == mpq(3,4)
    assert mpq(1,2) * 0 == mpq(0,1)
    assert mpq(1,2) * mpz(1) == mpq(1,2)
    assert mpq(1,2) * (-1) == mpq(-1,2)
    assert mpq(1,1) * mpc(1,0) == mpc('1.0+0.0j')
    assert mpc(1,0) * mpq(1,1) == mpc('1.0+0.0j')
    assert mpq(1,2) * z == mpq(1,1)
    assert mpq(1,2) * q == mpq(3,4)

    ctx = gmpy2.context()

    assert ctx.mul(mpq(1,2), mpq(3,2)) == mpq(3,4)
    assert ctx.mul(mpq(1,2), Fraction(3,2)) == mpq(3,4)
    assert ctx.mul(Fraction(1,2), mpq(3,2)) == mpq(3,4)
    assert ctx.mul(Fraction(1,2), Fraction(3,2)) == mpq(3,4)

    pytest.raises(TypeError, lambda: ctx.mul(1,'a'))
    pytest.raises(TypeError, lambda: mpq(1,2) * 'a')
    pytest.raises(TypeError, lambda: 'a' * mpq(1,2))

    a = mpq(3,11)
    b = mpq(1,2)

    assert a*b == mpq(3,22)
    assert b*a == mpq(3,22)
    assert a*0 == mpq(0,1)
    assert 0*a == mpq(0,1)
    assert a*-1 == mpq(-3,11)
    assert -1*a == mpq(-3,11)
    assert a*mpz(17) == mpq(51,11)
    assert mpz(17)*a == mpq(51,11)
    assert a*a == mpq(9,121)

    pytest.raises(TypeError, lambda: a*'b')
    pytest.raises(TypeError, lambda: 'b'*a)

    assert a * float('Inf') == mpfr('inf')
    assert float('Inf') * a == mpfr('inf')
    assert a * float('-Inf') == mpfr('-inf')
    assert float('-Inf') * a == mpfr('-inf')
    assert -a * float('Inf') == mpfr('-inf')
    assert float('Inf') * -a == mpfr('-inf')
    assert -a * float('-Inf') == mpfr('inf')
    assert float('-Inf') * -a == mpfr('inf')
    assert is_nan(a * float('nan'))
    assert is_nan(float('nan') * a)
    assert is_nan(mpz(0) * float('Inf'))
    assert is_nan(mpz(0) * float('-Inf'))
    assert is_nan(float('Inf') * mpz(0))
    assert is_nan(float('-Inf') * mpz(0))

    assert a * mpfr('Inf') == mpfr('inf')
    assert mpfr('Inf') * a == mpfr('inf')
    assert a * mpfr('-Inf') == mpfr('-inf')
    assert mpfr('-Inf') * a == mpfr('-inf')
    assert -a * mpfr('Inf') == mpfr('-inf')
    assert mpfr('Inf') * -a == mpfr('-inf')
    assert -a * mpfr('-Inf') == mpfr('inf')
    assert mpfr('-Inf') * -a == mpfr('inf')
    assert is_nan(a * mpfr('nan'))
    assert is_nan(mpfr('nan') * a)
    assert is_nan(mpz(0) * mpfr('Inf'))
    assert is_nan(mpz(0) * mpfr('-Inf'))
    assert is_nan(mpfr('Inf') * mpz(0))
    assert is_nan(mpfr('-Inf') * mpz(0))


def test_mpq_mod():
    a = mpq(3,11)
    b = mpq(1,2)
    ctx = gmpy2.get_context()

    assert a%b == mpq(3,11)
    assert b%a == mpq(5,22)
    assert a%z == mpq(3,11)
    assert mpq(3,1) % q == mpq(0,1)

    pytest.raises(ZeroDivisionError, lambda: a % mpq(0,5))

    assert ctx.mod(a, mpfr(1.5)) == mpfr('0.27272727272727271')

    pytest.raises(TypeError, lambda: ctx.mod(a, mpc(0.5, 2)))
    pytest.raises(ZeroDivisionError, lambda: ctx.mod(a, mpq(0, 2)))

    assert a % mpfr(1.5) == mpfr('0.27272727272727271')

    pytest.raises(TypeError, lambda: a % mpc(0.5, 2))
    pytest.raises(TypeError, lambda: a % 'str')
    pytest.raises(TypeError, lambda: ctx.mod(a, 'str'))


def test_mpq_divmod():
    a = mpq(3,11)
    b = mpq(1,2)
    c = Fraction(5,7)
    ctx = gmpy2.get_context()

    assert divmod(a,b) == (mpz(0), mpq(3,11))
    assert divmod(b,a) == (mpz(1), mpq(5,22))

    pytest.raises(ZeroDivisionError, lambda: divmod(a,0))

    assert divmod(17,a) == (mpz(62), mpq(1,11))
    assert divmod(mpz(17),a) == (mpz(62), mpq(1,11))
    assert divmod(a,c) == (mpz(0), mpq(3,11))

    pytest.raises(TypeError, lambda: divmod(mpq(1,2),'a'))

    ctx = gmpy2.ieee(64)
    gmpy2.set_context(ctx)

    assert ctx.divmod(mpq(3,2),mpq(3,7)) == (mpz(3), mpq(3,14))

    pytest.raises(TypeError, lambda: divmod(mpq(1,2), mpc(1,2)))

    assert divmod(a, float('Inf')) == (mpfr('0.0'), mpfr('0.27272727272727271'))
    assert divmod(a, float('-Inf')) == (mpfr('-1.0'), mpfr('-inf'))
    assert divmod(-a, float('Inf')) == (mpfr('-1.0'), mpfr('inf'))
    assert divmod(-a, float('-Inf')) == (mpfr('0.0'), mpfr('-0.27272727272727271'))
    assert all(map(is_nan, divmod(a, float('nan'))))
    assert all(map(is_nan, divmod(-a, float('nan'))))
    assert divmod(mpz(0), float('Inf')) == (mpfr('0.0'), mpfr('0.0'))
    assert divmod(mpz(0), float('-Inf')) == (mpfr('-0.0'), mpfr('-0.0'))
    assert all(map(is_nan, divmod(mpz(0), float('nan'))))
    assert all(map(is_nan, divmod(float('Inf'), a)))
    assert all(map(is_nan, divmod(float('-Inf'), a)))
    assert all(map(is_nan, divmod(float('Inf'), -a)))
    assert all(map(is_nan, divmod(float('-Inf'), -a)))
    assert all(map(is_nan, divmod(float('nan'), a)))
    assert all(map(is_nan, divmod(float('nan'), -a)))
    assert all(map(is_nan, divmod(float('Inf'), mpz(0))))
    assert all(map(is_nan, divmod(float('-Inf'), mpz(0))))
    assert all(map(is_nan, divmod(float('nan'), mpz(0))))

    assert divmod(a, mpfr('Inf')) == (mpfr('0.0'), mpfr('0.27272727272727271'))
    assert divmod(a, mpfr('-Inf')) == (mpfr('-1.0'), mpfr('-inf'))
    assert divmod(-a, mpfr('Inf')) == (mpfr('-1.0'), mpfr('inf'))
    assert divmod(-a, mpfr('-Inf')) == (mpfr('0.0'), mpfr('-0.27272727272727271'))
    assert all(map(is_nan, divmod(a, mpfr('nan'))))
    assert all(map(is_nan, divmod(-a, mpfr('nan'))))
    assert divmod(mpz(0), mpfr('Inf')) == (mpfr('0.0'), mpfr('0.0'))
    assert divmod(mpz(0), mpfr('-Inf')) == (mpfr('-0.0'), mpfr('-0.0'))
    assert divmod(mpz(0), mpfr('nan'))
    assert divmod(mpfr('Inf'), a)
    assert divmod(mpfr('-Inf'), a)
    assert divmod(mpfr('Inf'), -a)
    assert divmod(mpfr('-Inf'), -a)
    assert divmod(mpfr('nan'), a)
    assert divmod(mpfr('nan'), -a)
    assert divmod(mpfr('Inf'), mpz(0))
    assert divmod(mpfr('-Inf'), mpz(0))
    assert divmod(mpfr('nan'), mpz(0))


def test_mpq_floordiv():
    ctx = gmpy2.get_context()
    a, b = mpz(45), mpz(6)
    r, r2 = mpfr(45), mpfr(3.1)
    q, q2 = mpq(118,18), mpq(3,2)
    pyq, pyq2 = Fraction(118,18), Fraction(3,2)
    c, c2 = mpc(51, 65), mpc(4, 6)

    assert ctx.floor_div(q, q2) == mpz(4)
    assert ctx.floor_div(q, pyq2) == mpz(4)
    assert ctx.floor_div(q, pyq) == mpz(1)
    assert ctx.floor_div(pyq, q2) == mpz(4)
    assert ctx.floor_div(pyq, pyq2) == mpz(4)
    assert ctx.floor_div(pyq, q) == mpz(1)

    pytest.raises(ZeroDivisionError, lambda: ctx.floor_div(q, mpq(0,1)))
    pytest.raises(ZeroDivisionError, lambda: ctx.floor_div(q, Fraction(0)))
    pytest.raises(ZeroDivisionError, lambda: ctx.floor_div(pyq, mpq(0,1)))
    pytest.raises(ZeroDivisionError, lambda: ctx.floor_div(pyq, Fraction(0)))

    assert q // b == mpz(1)
    assert q // q2 == mpz(4)
    assert q // r2 == mpfr('2.0')

    pytest.raises(TypeError, lambda: q // c2)
    pytest.raises(TypeError, lambda: q // 'not')

    a = mpq(3,11)
    b = mpq(1,2)

    assert a//b == mpz(0)
    assert b//a == mpz(1)

    pytest.raises(ZeroDivisionError, lambda: a//0)

    assert 0//a == mpz(0)
    assert mpq(355, 113) // 2 == mpz(1)
    assert mpq(355, 113) // mpz(2) == mpz(1)
    assert mpq(355, 113) // z == mpz(1)


def test_mpq_truediv():
    a = mpq(3,11)
    b = mpq(1,2)
    ctx = gmpy2.get_context()

    assert a/b == mpq(6,11)
    assert gmpy2.div(a, b) == mpq(6,11)
    assert ctx.div(a, b) == mpq(6,11)
    assert b/a == mpq(11,6)

    pytest.raises(ZeroDivisionError, lambda: a/0)

    assert 0/a == mpq(0,1)
    assert a / z == mpq(3,22)
    assert mpq(3,11) / q == mpq(2,11)
    assert a / mpc(5, 4) == mpc('0.03325942350332594-0.02660753880266075j')

    pytest.raises(ZeroDivisionError, lambda: a / mpq(0,1))
    pytest.raises(TypeError, lambda: a / 'str')

    assert a / float('Inf') == mpfr('0.0')
    assert -a / float('Inf') == mpfr('-0.0')
    assert float('Inf') / a == mpfr('inf')
    assert float('Inf') / -a == mpfr('-inf')
    assert a / float('-Inf') == mpfr('-0.0')
    assert -a / float('-Inf') == mpfr('0.0')
    assert float('-Inf') / a == mpfr('-inf')
    assert float('-Inf') / -a == mpfr('inf')
    assert is_nan(a / float('nan'))
    assert is_nan(float('nan') / a)
    assert is_nan(float('nan') / mpz(0))

    assert a / mpfr('Inf') == mpfr('0.0')
    assert -a / mpfr('Inf') == mpfr('-0.0')
    assert mpfr('Inf') / a == mpfr('inf')
    assert mpfr('Inf') / -a == mpfr('-inf')
    assert a / mpfr('-Inf') == mpfr('-0.0')
    assert -a / mpfr('-Inf') == mpfr('0.0')
    assert mpfr('-Inf') / a == mpfr('-inf')
    assert mpfr('-Inf') / -a == mpfr('inf')
    assert is_nan(a / mpfr('nan'))
    assert is_nan(mpfr('nan') / a)
    assert is_nan(mpfr('nan') / mpz(0))
    assert is_nan(mpfr('nan') / mpz(0))


def test_mpq_pow():
    q = mpq(2,3)
    ctx = gmpy2.get_context()

    assert q ** 2 == mpq(4,9)
    assert q ** 0 == mpq(1,1)
    assert q ** -5 == mpq(243,32)
    assert ctx.pow(Fraction(2,3),2) == q ** 2
    assert mpq(-5,8) ** 5 == mpq(-3125,32768)
    assert q ** mpq(4,5) == mpfr('0.72298118079846574')

    pytest.raises(TypeError, lambda: pow(q, 5, 2))


def test_mpq_attributes():
    q = mpq('4/5')
    a = mpq(3,11)
    pyq = Fraction(4, 5)

    assert q.numerator == mpz(4)
    assert q.denominator == mpz(5)
    assert gmpy2.numer(q) == mpz(4)
    assert gmpy2.denom(q) == mpz(5)

    assert a.numerator == mpz(3)
    assert a.denominator == mpz(11)
    assert a.real == mpq(3,11)
    assert a.imag == mpz(0)

    pytest.raises(TypeError, lambda: gmpy2.numer(6.2))
    pytest.raises(TypeError, lambda: gmpy2.denom(5.6))
    pytest.raises(TypeError, lambda: gmpy2.denom(mpfr(5)))

    assert gmpy2.numer(pyq) == mpz(4)
    assert gmpy2.denom(pyq) == mpz(5)


def test_mpq_conjugate():
    a = mpq(3, 11)

    assert a.conjugate() == a


def test_mpq_floor():
    assert math.floor(mpq('7/2')) == mpz(3)


def test_mpq_ceil():
    assert math.ceil(mpq('7/2')) == mpz(4)


def test_mpq_trunc():
    assert math.trunc(mpq('7/2')) == mpz(3)


def test_mpq_not():
    q = mpq('4/5')

    assert q
    assert not mpq('0/5')


def test_mpq_qdiv():
    q = mpq('4/5')
    pyq = Fraction(4, 5)

    assert gmpy2.qdiv(q) == mpq(4,5)
    assert gmpy2.qdiv(pyq) == mpq(4,5)
    assert gmpy2.qdiv(5) == mpz(5)

    pytest.raises(TypeError, lambda: gmpy2.qdiv(mpc(4, 5)))
    pytest.raises(TypeError, lambda: gmpy2.qdiv(4, 5, 4))
    pytest.raises(TypeError, lambda: gmpy2.qdiv(4, 5.6))

    assert gmpy2.qdiv(q, 2) == mpq(2,5)
    assert gmpy2.qdiv(10, q) == mpq(25,2)
    assert gmpy2.qdiv(1) == mpz(1)

    assert gmpy2.qdiv(8,1) == mpz(8)
    assert gmpy2.qdiv(8,2) == mpz(4)
    assert gmpy2.qdiv(8,3) == mpq(8,3)
    assert gmpy2.qdiv(8,4) == mpz(2)
    assert gmpy2.qdiv(mpq(3,4), mpq(1,3)) == mpq(9,4)
    assert gmpy2.qdiv(mpq(3,4), mpq(1,4)) == mpz(3)

    args = mpq(2), 1/gmpy2.mpq(2)

    assert gmpy2.qdiv(*args) == mpz(4)
    assert args == (mpq(2), 1/mpq(2))


def test_mpq_repr():
    assert repr(mpq(11,13)) == 'mpq(11,13)'


def test_mpq_str():
    assert str(mpq(11,13)) == '11/13'


def test_issue_334():
    x = mpq(3,2)
    y = mpq(x,2)

    assert x == mpq(3,2)
    assert y == mpq(3,4)
    assert id(x) is not id(y)



from .gmpy2 import *
from .gmpy2 import __version__
# Internal variables/functions are not imported by * above.
# These are used by some python level functions and are needed
# at the top level.
# Use try...except to for static builds were _C_API is not available.
try:
    from .gmpy2 import _C_API, _mpmath_normalize, _mpmath_create
except ImportError:
    from .gmpy2 import _mpmath_normalize, _mpmath_create
